#include <cache/cache.h>

static Mutex cachesMutex;
static List  caches;

Bool
__cache_enter__ (void)
{
   mutex_create (& cachesMutex);

   list_clear (& caches);

   return TRUE;
}

Bool
cache_create (Cache * cache,
              Size objectSize)
{
   mutex_lock (& cachesMutex);

   if (objectSize < sizeof (Object))
     objectSize = sizeof (Object);

   mutex_create (& cache -> lock);

   list_clear (& cache -> slabsUsed);
   list_clear (& cache -> slabsFree);

   cache -> slabNotEmpty = NULL;
   cache -> objectSize   = objectSize;

   list_insert_front (& caches, cache);

   mutex_unlock (& cachesMutex);

   return TRUE;
}

void
cache_destroy (Cache * cache)
{
   mutex_lock (& cache -> lock);

   cache_flush (cache, ~0);

   while (list_empty (& cache -> slabsUsed) == FALSE)
     slab_destroy ((Slab *) list_front (& cache -> slabsUsed));

   list_remove (& cache -> cacheList);

   mutex_unlock (& cache -> lock);

   mutex_destroy (& cache -> lock);
}

static Slab *
cache_find (Cache * cache)
{
   ListIterator position;
   
   if (list_empty (& cache -> slabsFree) == FALSE)
   {
      Slab * slab = (Slab *) list_remove_front (& cache -> slabsFree);
      
      list_insert_front (& cache -> slabsUsed, slab);
      
      cache -> slabNotEmpty = slab;
      
      return slab;
   }

   for (position = list_begin (& cache -> slabsUsed);
        position != list_end (& cache -> slabsUsed);
	position = list_next (position))
   {
      Slab * slab = (Slab *) position;

      if (list_empty (& slab -> objectsFree) == FALSE)
      {
         cache -> slabNotEmpty = slab;
      
         return slab;
      }
   }

   return NULL;
}

void *
cache_allocate (Cache * cache)
{
   Slab * slab;
   Object * object;

   mutex_lock (& cache -> lock);

   slab = cache -> slabNotEmpty;
  
   if (slab == NULL ||
       list_empty (& slab -> objectsFree) == TRUE)
   {
       slab = cache_find (cache);
   
       if (slab == NULL)
       {
          if (slab_create (& slab, cache) == FALSE)
            return NULL;
       }
   }

   object = (Object *) list_remove_front (& slab -> objectsFree);
   ++ slab -> objectsAllocatedCount;

   list_move_to_back (& caches, (ListIterator) cache);

   mutex_unlock (& cache -> lock);

   return object;
}

void
cache_free (Cache * cache,
            void  * object)
{
     Slab * slab = (Slab *) ((Size) object & ~(SLAB_SIZE - 1)),
          * checkSlab;

     if (slab -> cache != cache ||
	 object < (void *) & slab [1] ||
	 (((uint8 *) object - (uint8 *) & slab [1]) % cache -> objectSize) > 0)
       return;

     mutex_lock (& cache -> lock);

     checkSlab = (Slab *) ((Size) list_next ((ListIterator) object) & ~(SLAB_SIZE - 1));
     if (checkSlab == slab)
     {
	ListIterator currentObject;

	for (currentObject = list_begin (& slab -> objectsFree);
	     currentObject != list_end (& slab -> objectsFree);
	     currentObject = list_next (currentObject))
          if ((void *) currentObject == object)
	  {
             mutex_unlock (& cache -> lock);
             
             return;
          }
     }
     
     list_insert_front (& slab -> objectsFree, object); 
     
     -- slab -> objectsAllocatedCount;
     if (slab -> objectsAllocatedCount <= 0)
     {
        list_remove (& slab -> slabList);
	list_insert_front (& cache -> slabsFree, & slab -> slabList);

        if (cache -> slabNotEmpty == slab)
          cache -> slabNotEmpty = NULL;
     }
     else
     if (cache -> slabNotEmpty == NULL ||
         slab -> objectsAllocatedCount >= cache -> slabNotEmpty -> objectsAllocatedCount)
       cache -> slabNotEmpty = slab;

     mutex_unlock (& cache -> lock);
}

Size
cache_flush (Cache * cache, Size limit)
{
    Size result;

    mutex_lock (& cache -> lock);

    for (result = 0;
         list_empty (& cache -> slabsFree) == FALSE &&
	    result < limit;
	 result += SLAB_SIZE)
      slab_destroy ((Slab *) list_front (& cache -> slabsFree));

    mutex_unlock (& cache -> lock);

    return result;
}

Size
cache_reap (Size limit)
{
    Size result;
    ListIterator cache;
   
    mutex_lock (& cachesMutex);

    for (result   = 0,
	    cache = list_begin (& caches);
	 cache != list_end (& caches) &&
	    result < limit;
	 cache = list_next (cache))
      result += cache_flush ((Cache *) cache, limit - result);

    mutex_unlock (& cachesMutex);

    return result;
}

