#include <utility/memory.h>
#include <utility/string.h>
#include <memory/memory.h>
#include <scheduler/mutex.h>
#include <console/type.h>
#include <console/console.h>

static Mutex consoleMutex;
static Console * consoleCurrent = NULL;
static uint8 consoleDefault [sizeof (Console) + 80 * 25 * sizeof (uint16)];

Bool
__console_enter__ (void)
{
        mutex_create (& consoleMutex);

        console_create (NULL,
                        CONSOLE_TYPE_DEFAULT,
                        80,
                        25);
        
        return TRUE;
}

Bool
console_create (Console ** result,
                ConsoleType * type,
                ConsoleCoordinate xVirtualSize,
                ConsoleCoordinate yVirtualSize)
{
        Console * console;
        
        if (consoleCurrent == NULL)
          console = (Console *) consoleDefault;
        else
          console = (Console *) memory_allocate (sizeof (Console) +
                                                 xVirtualSize * yVirtualSize * sizeof (uint16));
        
        if (console == NULL)
          return FALSE;

        if (type == NULL)
          type = CONSOLE_TYPE_DEFAULT;

        console -> type            = type;
        console -> foregroundColor = CONSOLE_COLOR_WHITE;
        console -> backgroundColor = CONSOLE_COLOR_BLACK;
        console -> xCursor         = 0;
        console -> yCursor         = 0;
        console -> xSize           = 0;
        console -> ySize           = 0;
        console -> xVirtualOrigin  = 0;
        console -> yVirtualOrigin  = 0;
        console -> xVirtualSize    = xVirtualSize;
        console -> yVirtualSize    = yVirtualSize;
        console -> textMemory      = (uint16 *) & console [1];

        memory_zero (console -> textMemory, xVirtualSize * yVirtualSize * sizeof (uint16));

        if ((* console -> type -> create) (console, xVirtualSize, yVirtualSize) == FALSE)
        {
           memory_free (console);

           return FALSE;
        }

        if (result != NULL)
          * result = console;
        else
          console_activate (console);

        return TRUE;
}

void
console_destroy (Console * console)
{
        mutex_lock (& consoleMutex);

        if (console == consoleCurrent)
          console_activate ((Console *) consoleDefault);

        if (console != (Console *) consoleDefault)
          memory_free (console);

        mutex_unlock (& consoleMutex);
}

Console *
console_current (void)
{
        return consoleCurrent;
}

void
console_activate (Console * console)
{
        mutex_lock (& consoleMutex);

        if (consoleCurrent != NULL)
          (* consoleCurrent -> type -> deactivate) (consoleCurrent);

        consoleCurrent = console;

        (* console -> type -> activate) (console);
        (* console -> type -> setCursor) (console, console -> xCursor, console -> yCursor);
        
        console_render (console);

        mutex_unlock (& consoleMutex);
}

void
console_set_cursor (Console * console, ConsoleCoordinate x, ConsoleCoordinate y)
{
        mutex_lock (& consoleMutex);

        if (console == NULL)
          console = consoleCurrent;

        console -> xCursor = x;
        console -> yCursor = y;

        if (console == consoleCurrent)
          (* console -> type -> setCursor) (console, x, y);

        mutex_unlock (& consoleMutex);
}

void
console_set_origin (Console * console, ConsoleCoordinate x, ConsoleCoordinate y)
{
        mutex_lock (& consoleMutex);

        if (console == NULL)
          console = consoleCurrent;

        console -> xVirtualOrigin = x;
        console -> yVirtualOrigin = y;

        if (console == consoleCurrent)
          console_render (console);

        mutex_unlock (& consoleMutex);
}

void
console_set_foreground_color (Console * console, ConsoleColor foregroundColor)
{
        mutex_lock (& consoleMutex);

        if (console == NULL)
          console = consoleCurrent;

        console -> foregroundColor = foregroundColor;

        mutex_unlock (& consoleMutex);
}

void
console_set_background_color (Console * console, ConsoleColor backgroundColor)
{
        mutex_lock (& consoleMutex);

        if (console == NULL)
          console = consoleCurrent;

        console -> backgroundColor = backgroundColor;

        mutex_unlock (& consoleMutex);
}

void
console_render (Console * console)
{
        const uint16 * line;
        Size lines, lineWidth;

        mutex_lock (& consoleMutex);

        if (console == NULL)
          console = consoleCurrent;

        lineWidth = console -> xVirtualSize - console -> xVirtualOrigin;

        if (lineWidth > console -> xSize)
          lineWidth = console -> xSize;

        for (line = console -> textMemory + 
                    console -> yVirtualOrigin * console -> xVirtualSize + 
                     console -> xVirtualOrigin,
               lines = console -> ySize;
             lines > 0;
             line += console -> xVirtualSize,
               -- lines)
        {
           if (line >= & console -> textMemory [console -> xVirtualSize * console -> yVirtualSize])
             line -= console -> xVirtualSize * console -> yVirtualSize;

           (* console -> type -> renderLine) (console, 
                                              console -> ySize - lines, 
                                              lineWidth, 
                                              line);
        }

        mutex_unlock (& consoleMutex);
}

void
console_newline (Console * console)
{
        mutex_lock (& consoleMutex);

        if (console == NULL)
          console = consoleCurrent;

        if (console -> yCursor >= console -> ySize - 1)
        {
           if (console -> yVirtualOrigin >= console -> yVirtualSize - 1)
             console -> yVirtualOrigin = 0;
           else
             ++ console -> yVirtualOrigin;
        }
        else
          ++ console -> yCursor;

        console -> xVirtualOrigin = 0;
        console -> xCursor        = 0;

        memory_zero (console -> textMemory +
                     ((console -> yVirtualOrigin + console -> yCursor) %
                      console -> yVirtualSize) * console -> xVirtualSize,
                     console -> xVirtualSize * sizeof (uint16));

        mutex_unlock (& consoleMutex);
}

void
console_tab (Console * console)
{
        Size tabSpaces;

        mutex_lock (& consoleMutex);

        if (console == NULL)
          console = consoleCurrent;

        tabSpaces = 8 - (console -> xCursor % 8);
        
        if (console -> xCursor + tabSpaces >= console -> xVirtualSize)
          tabSpaces = console -> xVirtualSize - console -> xCursor - 1;

        memory_zero (console -> textMemory +
                        ((console -> yVirtualOrigin + console -> yCursor) %
                          console -> yVirtualSize) * console -> xVirtualSize + 
                        console -> xVirtualOrigin + console -> xCursor,
                     tabSpaces * sizeof (uint16));

        console -> xCursor += tabSpaces;
        
        if (console -> xCursor >= console -> xSize)
        {
           console -> xVirtualOrigin += (console -> xCursor - console -> xSize) + 1;
           console -> xCursor         = console -> xSize - 1;
        }

        mutex_unlock (& consoleMutex);
}

void
console_print (Console * console, const char * text)
{
        uint16 * lineStart,
               * lineCurrent;
        Size lineOccupied;
        uint16 color;
        
        mutex_lock (& consoleMutex);

        if (console == NULL)
          console = consoleCurrent;

        color = (console -> backgroundColor << 12) | (console -> foregroundColor << 8);

        while (* text != '\0')
        {
           lineStart = console -> textMemory + 
                          ((console -> yVirtualOrigin + console -> yCursor) % console -> yVirtualSize) * 
                           console -> xVirtualSize;

           lineCurrent = lineStart + console -> xVirtualOrigin + console -> xCursor;

           while (* text >= ' ' &&
                  lineCurrent - lineStart < console -> xVirtualSize)
           {
               * lineCurrent = * (uint8 *) text | color;

               ++ lineCurrent;
               ++ text;
           }

           while (* text >= ' ')
             ++ text;

           lineOccupied = lineCurrent - lineStart;

           if (lineOccupied >= console -> xVirtualOrigin + console -> xSize)
           {
              console -> xCursor        = console -> xSize - 1;
              console -> xVirtualOrigin = lineOccupied - console -> xSize + 1;
           }
           else
              console -> xCursor = lineOccupied - console -> xVirtualOrigin;

           while (* text < ' ' &&
                  * text != '\0')
           {
              switch (* text)
              {
              case '\n':
                   console_newline (console);

                   break;

              case '\t':
                   console_tab (console);

                   break;

              default:
                   break;
              }

              ++ text;
           }
        } 

        if (console == consoleCurrent)
        {
           (* console -> type -> setCursor) (console, console -> xCursor, console -> yCursor);
           
           console_render (console);
        }

        mutex_unlock (& consoleMutex);
}

void
console_print_arguments (Console * console, const char * format, Arguments arguments)
{
        static char formatBuffer [4096];

        mutex_lock (& consoleMutex);

        string_format_arguments (formatBuffer, sizeof (formatBuffer), format, arguments);

        console_print (console, formatBuffer);

        mutex_unlock (& consoleMutex);
}

void
console_print_format (Console * console, const char * format, ...)
{
        Arguments arguments;

        ARGUMENTS_BEGIN (arguments, format);

        console_print_arguments (console, format, arguments);

        ARGUMENTS_END (arguments);
}

