#include <utility/list.h>
#include <utility/string.h>
#include <utility/memory.h>
#include <console/console.h>
#include <scheduler/thread.h>
#include <linker/component.h>
#include <linker/interface.h>
#include <linker/section.h>
#include <linker/export.h>
#include <linker/linker.h>
#include <libc/stdlib.h>
#include <libc/stdio.h>
#include <debugger/pong.h>
#include <debugger/breakpoint.h>
#include <debugger/command.h>

Bool
command_help (Size argumentCount, char ** arguments)
{
    console_print (NULL,
                   "help - print help information\n"
                   "pong - when in doubt, play pong\n"
                   "components - list components and their interfaces\n"
                   "component name - list component's interfaces and sections\n"
                   "section component name - list section in component\n"
                   "export interface name - list export in interface\n"
                   "break component function - set a breakpoint at the function in component\n"
                   "enable component function - enable the specified breakpoint\n"
                   "disable component function - disable the specified breakpoint\n"
                   "run command - run the command through libc:system ()\n");
    
    return TRUE;
}

Bool
command_components (Size argumentCount, char ** arguments)
{
    Linker * linker = linker_current ();
    ListIterator componentPos;
    Bool noisy = argumentCount >= 1 && string_to_integer (arguments[0]) != 0;

    for (componentPos = list_begin (& linker -> components);
         componentPos != list_end (& linker -> components);
         componentPos = list_next (componentPos))
    {
        Component * component = (Component *) componentPos;
        ListIterator interfacePos;

        console_print_format (NULL, "component: %s\n", component -> name);

        if (noisy)
        for (interfacePos = list_begin (& component -> interfaces);
             interfacePos != list_end (& component -> interfaces);
             interfacePos = list_next (interfacePos))
        {
            Interface * interface = (Interface *) interfacePos;
                  
            console_print_format (NULL, "  interface: %s\n", interface -> name);
        }
    }

    return TRUE;
}

Bool
command_component (Size argumentCount, char ** arguments)
{
    Component * component;
    ListIterator interfacePos,
                 sectionPos;
    Bool noisy = argumentCount >= 2 && string_to_integer (arguments[1]) != 0;
    
    if (argumentCount < 1)
      return FALSE;
    
    component = linker_find_component (arguments [0]);

    if (component == NULL)
    {
       console_print_format (NULL, "Unknown component %s!\n", arguments [0]);

       return TRUE;
    }

    if (noisy)
      console_print_format (NULL,
                          "component: %s\n"
                          "  imports: %u\n"
                          "  constructors: %u\n"
                          "  destructors: %u\n"
                          "  references: %u\n",
                          component -> name,
                          list_size (& component -> imports),
                          list_size (& component -> constructors),
                          list_size (& component -> destructors),
                          component -> referenceCount);
    else
      console_print_format (NULL, "component: %s\n", component -> name);

    for (interfacePos = list_begin (& component -> interfaces);
         interfacePos != list_end (& component -> interfaces);
         interfacePos = list_next (interfacePos))
    {
        Interface * interface = (Interface *) interfacePos;

        console_print_format (NULL, "  interface: %s\n", interface -> name);
    }

    if (noisy)
    for (sectionPos = list_begin (& component -> sections);
         sectionPos != list_end (& component -> sections);
         sectionPos = list_next (sectionPos))
    {
        Section * section = (Section *) sectionPos;

        console_print_format (NULL, "  section: %s\n", section -> name);
    }
           
    return TRUE;
}

Bool
command_section (Size argumentCount, char ** arguments)
{
     Component * component;
     Section * section;

     if (argumentCount < 2)
       return FALSE;
          
     component = linker_find_component (arguments [0]);

     if (component == NULL)
     {
        console_print_format (NULL, "Unknown component %s!\n", arguments [0]);

        return TRUE;
     }
         
     section = component_find_section (component, arguments [1]);

     if (section == NULL)
     {
        console_print_format (NULL, "Unknown section %s:%s!\n", arguments [0], arguments [1]);

        return TRUE;
     }

     console_print_format (NULL,
                           "section: %s\n"
                           "  address: 0x%x\n"
                           "  alignment: %u, 0x%x\n"
                           "  size: %u, 0x%x\n"
                           "  capabilities: %s%s%s\n"
                           "  relocations: %u\n"
                           "  exports: %u\n",
                           section -> name,
                           section -> data,
                           section -> dataAlignment,
                           section -> dataAlignment,
                           section -> dataSize,
                           section -> dataSize,
                           section -> capabilities & SECTION_CAPABILITY_READ ? "r" : "",
                           section -> capabilities & SECTION_CAPABILITY_WRITE ? "w" : "",
                           section -> capabilities & SECTION_CAPABILITY_EXECUTE ? "x" : "",
                           section -> relocationCount,
                           list_size (& section -> exports));

     return TRUE;
}

Bool
command_export (Size argumentCount, char ** arguments)
{
    Export * export;

    if (argumentCount < 2)
      return FALSE;

    export = linker_find_export (arguments [0], arguments [1]);

    if (export == NULL)
    {
       console_print_format (NULL,
                             "Unknown export %s:%s!\n\n",
                             arguments [0],
                             arguments [1]);

       return TRUE;
    }

    console_print_format (NULL,
                          "export: %s\n"
                          "  interface: %s\n"
                          "  offset: %u, 0x%x\n"
                          "  size: %u, 0x%x\n"
                          "  type: %s\n"
                          "  address: 0x%x\n"
                          "  section: %s\n"
                          "  component: %s\n",
                          export -> name,
                          export -> interface -> name,
                          export -> offset,
                          export -> offset,
                          export -> size,
                          export -> size,
                          export -> type == EXPORT_TYPE_UNKNOWN ? "unknown" :
                          export -> type == EXPORT_TYPE_DATA ? "data" :
                          export -> type == EXPORT_TYPE_FUNCTION ? "function" : "invalid",
                          export -> section -> data + export -> offset,
                          export -> section -> name,
                          export -> section -> component -> name);

    return TRUE;
}

Bool
command_break (Size argumentCount, char ** arguments)
{
    BreakPoint * breakPoint;

    if (argumentCount < 2)
      return FALSE;

    if (breakpoint_find_function (arguments [0], arguments [1]) != NULL)
    {
       console_print_format (NULL,
                             "Breakpoint %s:%s at 0x%x already exists!\n\n",
                             arguments [0],
                             arguments [1],
                             breakPoint -> address);

       return TRUE;
    }

    if (breakpoint_create (& breakPoint, arguments [0], arguments [1]) == FALSE)
    {
       console_print_format (NULL, 
                             "Failed to create breakpoint %s:%s!\n\n", 
                             arguments [0], 
                             arguments [1]);
       
       return TRUE;
    }

    console_print_format (NULL, 
                          "Breakpoint %s:%s at 0x%x.\n", 
                          arguments [0],
                          arguments [1],
                          breakPoint -> address);

    return TRUE;
}

Bool
command_enable (Size argumentCount, char ** arguments)
{
    BreakPoint * breakPoint;
          
    if (argumentCount < 2)
      return FALSE;
          
    breakPoint = breakpoint_find_function (arguments [0], arguments [1]);

    if (breakPoint == NULL)
    {
       console_print_format (NULL,
                             "Failed to enable breakpoint %s:%s\n",
                             arguments [0],
                             arguments [1]);
             
       return TRUE;
    }

    breakpoint_enable (breakPoint);

    console_print_format (NULL,
                          "Enabled breakpoint %s:%s.\n",
                          arguments [0],
                          arguments [1]);

    return TRUE;
}

Bool
command_disable (Size argumentCount, char ** arguments)
{
    BreakPoint * breakPoint;

    if (argumentCount < 2)
      return FALSE;

    breakPoint = breakpoint_find_function (arguments [0], arguments [1]);

    if (breakPoint == NULL)
    {
       console_print_format (NULL,
                             "Failed to disable breakpoint %s:%s\n",
                             arguments [0],
                             arguments [1]);

       return TRUE;
    }

    breakpoint_disable (breakPoint);

    console_print_format (NULL,
                          "Disabled breakpoint %s:%s.\n",
                          arguments [0],
                          arguments [1]);

    return TRUE;
}

static void
runThread (void * command)
{
    int exitStatus = system ((char *) command);

    console_print_format (NULL,
                          "Command exited with status %d.\n",
                          exitStatus);
}


Bool
command_run (Size argumentCount, char ** arguments)
{
    static char buffer [4096];

    char * bufferPosition;
    char ** argumentCurrent;

    if (argumentCount < 1)
      return FALSE;
    
    for (argumentCurrent = arguments,
            bufferPosition = buffer;
         argumentCurrent < & arguments [argumentCount] &&
            bufferPosition < & buffer [sizeof (buffer)];
         ++ argumentCurrent)
    {
       Size length = string_length (* argumentCurrent);

       if (length > & buffer [sizeof (buffer)] - bufferPosition)
         length = & buffer [sizeof (buffer)] - bufferPosition - 1;

       memory_copy (bufferPosition, * argumentCurrent, length);  

       bufferPosition += length;

       * bufferPosition ++ = ' ';
    }

    bufferPosition [-1] = '\0';

    if (thread_create (NULL,
                       THREAD_PRIORITY_MINIMUM,
                       THREAD_PRIORITY_MAXIMUM,
                       runThread,
                       buffer,
                       NULL,
                       0) == FALSE)
      console_print_format (NULL, "Failed to run command: %s\n", buffer); 

    return TRUE;
}

Bool
command_pong (Size argumentCount, char ** arguments)
{
     PongArena arena;

     if (pong_arena_create (& arena, 320, 200, 80, 20, 10, 10, 7) == FALSE)
     {
        console_print (NULL, "Pong failed! Panic!\n");

        return TRUE;
     }

     console_print (NULL,
                    "POS Pong!\n"
                    "q - quit\n"
                    "a - add ball\n"
                    "r - remove ball\n"
                    "cursor up - paddle up\n"
                    "cursor down - paddle down\n"
                    "cursor left - paddle left\n"
                    "cursor right - paddle right\n\n"
                    "Press a key to begin...\n");

     getchar ();

     pong_arena_play (& arena);

     pong_arena_destroy (& arena);

     return TRUE;
}

