#include <scheduler/critical.h>
#include <ide/controller.h>

static IDEController ideControllers [IDE_CONTROLLER_COUNT] =
{
        { 
          FALSE, 
          FALSE,
          IDE_CONTROLLER_NUMBER_0,
          IDE_CONTROLLER_IO_COMMAND_BASE_0,
          IDE_CONTROLLER_IO_CONTROL_BASE_0,
          IDE_CONTROLLER_IRQ_0
        },
        {
          FALSE,
          FALSE,
          IDE_CONTROLLER_NUMBER_1,
          IDE_CONTROLLER_IO_COMMAND_BASE_1,
          IDE_CONTROLLER_IO_CONTROL_BASE_1,
          IDE_CONTROLLER_IRQ_1
        }
};

Bool
__ide_enter__ (void)
{
        IDEControllerNumber controllerNumber;

        for (controllerNumber = 0;
             controllerNumber < IDE_CONTROLLER_COUNT;
             ++ controllerNumber)
          ide_controller_detect (controllerNumber);

        return TRUE;
}

static void
ide_controller_interrupt_handler (InterruptNumber number, void * data)
{
        IDEController * controller = (IDEController *) data;

        if (ide_controller_read_command_register_8 (controller, IDE_COMMAND_REGISTER_STATUS) & IDE_STATUS_COMMAND_ERROR)
          controller -> commandError = ide_controller_read_command_register_8 (controller, IDE_COMMAND_REGISTER_ERROR);

        if (controller -> awaitingAcknowledgement == TRUE)
        {
           controller -> awaitingAcknowledgement = FALSE;

           timer_stop (& controller -> commandTimer, NULL);
        
           semaphore_up (& controller -> commandAcknowledged);
        }
        else
        if (controller -> awaitingSectors > 0)
        {
           -- controller -> awaitingSectors;
           
           timer_reset (& controller -> commandTimer, NULL);

           if (controller -> commandError ||
               controller -> awaitingSectors <= 0)
           {
               controller -> awaitingSectors = 0;

               timer_stop (& controller -> commandTimer, NULL);

               semaphore_up (& controller -> commandAcknowledged);
           }
        }
}

static void
ide_controller_timeout_handler (Timer * timer, void * data)
{
        ide_controller_cancel_command ((IDEController *) data);
}

IDEController *
ide_controller_detect (IDEControllerNumber controllerNumber)
{
        IDEController * controller = & ideControllers [controllerNumber];
        IDEDriveNumber driveNumber;
        Size driveCount;
        TimeOffset timeout; 
       
        /* For now, just assume the controller exists. */
        timeout.nanoSeconds = IDE_CONTROLLER_COMMAND_TIMEOUT;
        timeout.seconds     = 0;

        timer_create (& controller -> commandTimer, 
                      & timeout, 
                      ide_controller_timeout_handler, 
                      controller, 
                      FALSE);

        semaphore_create (& controller -> commandAcknowledged, 0);

        interrupt_handler_create (& controller -> interruptHandler,
                                  controller -> irq,
                                  ide_controller_interrupt_handler,
                                  controller,
                                  0,
                                  "ide");
                                                                   
        ide_controller_write_control_register_8 (controller, 
                                                 IDE_CONTROL_REGISTER_DEVICE_CONTROL, 
                                                 0x08);

        for (driveNumber = 0,
                driveCount = 0;
             driveNumber < IDE_DRIVE_COUNT;
             ++ driveNumber)
          if (ide_drive_detect (controller, driveNumber) != NULL)
            ++ driveCount;

        if (driveNumber == 0)
        {
           timer_destroy (& controller -> commandTimer);
           semaphore_destroy (& controller -> commandAcknowledged);
           interrupt_handler_destroy (& controller -> interruptHandler);

           return NULL;
        }

        controller -> isDetected = TRUE;

        return controller;
}

IDEController *
ide_controller_allocate (IDEControllerNumber number)
{
        if (ideControllers [number].isAllocated == TRUE ||
            ideControllers [number].isDetected == FALSE)
          return NULL;

        ideControllers [number].isAllocated = TRUE;

        return & ideControllers [number];
}

void
ide_controller_free (IDEController * controller)
{
        controller -> isAllocated = FALSE;
}

void
ide_controller_issue_command (IDEController * controller,
                              IDEDriveNumber driveNumber,
                              IDECommand command,
                              uint16 cylinder,
                              uint8 head,
                              uint8 sector,
                              uint8 sectorCount)
{
        critical_enter ();
        
        controller -> commandError = IDE_ERROR_NONE;

        timer_start (& controller -> commandTimer);

        ide_controller_write_command_register_8 (controller,
                                                 IDE_COMMAND_REGISTER_SECTOR_COUNT,
                                                 sectorCount);
        ide_controller_write_command_register_8 (controller,
                                                 IDE_COMMAND_REGISTER_SECTOR_NUMBER,
                                                 sector);
        ide_controller_write_command_register_8 (controller,
                                                 IDE_COMMAND_REGISTER_CYLINDER_LOW,
                                                 cylinder & 0xFF);
        ide_controller_write_command_register_8 (controller,
                                                 IDE_COMMAND_REGISTER_CYLINDER_HIGH,
                                                 cylinder >> 8);
        ide_controller_write_command_register_8 (controller,
                                                 IDE_COMMAND_REGISTER_DRIVE_HEAD,
                                                 0xA0 | (driveNumber << 4) | head);
        ide_controller_write_command_register_8 (controller,
                                                 IDE_COMMAND_REGISTER_COMMAND,
                                                 command);

        critical_exit ();
}

void
ide_controller_poll (const IDEController * controller,
                     uint8 statusMask,
                     uint8 desiredStatus)
{
        uint8 status;

        do status = ide_controller_read_control_register_8 (controller, IDE_CONTROL_REGISTER_ALTERNATE_STATUS);
        while ((status & statusMask) != desiredStatus);
}

void
ide_controller_cancel_command (IDEController * controller)
{
        critical_enter ();

        timer_stop (& controller -> commandTimer, NULL);

        controller -> commandError = IDE_ERROR_COMMAND_ABORTED;

        if (controller -> awaitingAcknowledgement == TRUE)
        {
           controller -> awaitingAcknowledgement = FALSE;

           semaphore_up (& controller -> commandAcknowledged);
        }
        else
        if (controller -> awaitingSectors > 0)
        {
           controller -> awaitingSectors = 0;

           semaphore_up (& controller -> commandAcknowledged);
        }

        critical_exit ();
}

