#include <ide/controller.h>

void
ide_drive_read_buffer (const IDEDrive * drive,
                       void * buffer, 
                       Size length)
{
        if (drive -> has32BitIO == TRUE)
        {
           uint32 * currentPosition;

           for (currentPosition = (uint32 *) buffer,
                   length /= 4;
                length > 0;
                ++ currentPosition,
                   -- length)
              * currentPosition = ide_controller_read_command_register_32 (drive -> controller,
                                                                           IDE_COMMAND_REGISTER_DATA);
        }
        else
        {
           uint16 * currentPosition;

           for (currentPosition = (uint16 *) buffer,
                   length /= 2;
                length > 0;
                ++ currentPosition, 
                -- length)
             * currentPosition = ide_controller_read_command_register_16 (drive -> controller,
                                                                          IDE_COMMAND_REGISTER_DATA);
        }
}

void
ide_drive_write_buffer (const IDEDrive * drive,
                        const void * buffer,
                        Size length)
{
        if (drive -> has32BitIO == TRUE)
        {
           const uint32 * currentPosition;

           for (currentPosition = (uint32 *) buffer,
                   length /= 4;
                length > 0;
                ++ currentPosition,
                   -- length)
             ide_controller_write_command_register_32 (drive -> controller,
                                                       IDE_COMMAND_REGISTER_DATA,
                                                       * currentPosition);
        }
        else
        {
           const uint16 * currentPosition;
       
           for (currentPosition = (const uint16 *) buffer,
                   length /= 2;
                length > 0;
                ++ currentPosition,
                   -- length)
             ide_controller_write_command_register_16 (drive -> controller,
                                                       IDE_COMMAND_REGISTER_DATA,
                                                       * currentPosition);
        }
}

IDEError
ide_drive_recalibrate (const IDEDrive * drive)
{
        IDEController * controller = drive -> controller;
        
        controller -> awaitingAcknowledgement = TRUE;
       
        ide_controller_issue_command (controller,
                                      drive -> number,
                                      IDE_COMMAND_RECALIBRATE_DRIVE,
                                      0,
                                      0,
                                      1,
                                      1);

        semaphore_down (& controller -> commandAcknowledged);
     
        return controller -> commandError;
}

IDEError
ide_drive_identify (IDEDrive * drive)
{
        IDEController * controller = drive -> controller;

        controller -> awaitingAcknowledgement = TRUE;

        ide_controller_issue_command (controller,
                                      drive -> number,
                                      IDE_COMMAND_IDENTIFY_DRIVE,
                                      0,
                                      0,
                                      1,
                                      1);

        semaphore_down (& controller -> commandAcknowledged);

        if (controller -> commandError != IDE_ERROR_NONE)
          return controller -> commandError;

        ide_drive_read_buffer (drive,
                               & drive -> identification, 
                               sizeof (IDEDriveIdentification));

        return IDE_ERROR_NONE;
}

IDEError
ide_drive_read (const IDEDrive * drive,
                uint16 cylinder,
                uint8 head,
                uint8 sector,
                void * buffer,
                Size sectorCount)
{
        IDEController * controller = drive -> controller;
        Size transferSize;
        uint8 * currentPosition;

        for (currentPosition = (uint8 *) buffer;
             sectorCount > 0;
             currentPosition += transferSize * 512,
                sectorCount -= transferSize)
        {
                if (sectorCount > 256)
                  transferSize = 256;
                else
                  transferSize = sectorCount;

                controller -> awaitingSectors = transferSize;

                ide_controller_issue_command (controller,
                                              drive -> number,
                                              IDE_COMMAND_READ_SECTORS_WITH_RETRY,
                                              cylinder,
                                              head,
                                              sector,
                                              transferSize);
        
                semaphore_down (& controller -> commandAcknowledged);
                
                if (controller -> commandError != IDE_ERROR_NONE)
                  return controller -> commandError;

                ide_drive_read_buffer (drive, currentPosition, transferSize * 512);
        }

        return IDE_ERROR_NONE;
}

IDEError
ide_drive_write (const IDEDrive * drive,
                 uint16 cylinder,
                 uint8 head,
                 uint8 sector,
                 const void * buffer,
                 Size sectorCount)
{
        IDEController * controller = drive -> controller;
        Size transferSize;
        const uint8 * currentPosition;

        for (currentPosition = (uint8 *) buffer;
             sectorCount > 0;
             currentPosition += transferSize * 512,
                sectorCount -= transferSize)
        {
                if (sectorCount > 256)
                  transferSize = 256;
                else
                  transferSize = sectorCount;

                controller -> awaitingSectors = transferSize;

                ide_controller_issue_command (controller,
                                              drive -> number,
                                              IDE_COMMAND_WRITE_SECTORS_WITH_RETRY,
                                              cylinder,
                                              head,
                                              sector,
                                              transferSize);
        
                if (controller -> commandError != IDE_ERROR_NONE)
                  return controller -> commandError;

                ide_drive_write_buffer (drive, currentPosition, transferSize * 512);

                semaphore_down (& controller -> commandAcknowledged);

                if (controller -> commandError != IDE_ERROR_NONE)
                  return controller -> commandError;
        }

        return IDE_ERROR_NONE;
}

