#include <utility/memory.h>
#include <console/console.h>
#include <interrupt/interrupt.h>
#include <scheduler/semaphore.h>
#include <scheduler/thread.h>
#include <kbc/kbc.h>
#include <kbc/keyboard.h>

static Keyboard          keyboardDefault;
static Semaphore         keyboardSemaphore;
static Thread *          keyboardThread = NULL;
static Keyboard *        keyboardCurrent = NULL;
static InterruptHandler  keyboardHandler;

void
keyboard_interrupt_handler (InterruptNumber number, void * data)
{
        semaphore_up (& keyboardSemaphore);
}

void
keyboardEventDispatcher (void * data)
{
        for (;;)
        {  
           KeyModifier modifier = KEY_MODIFIER_NONE;
           uint8 led = KEYBOARD_LED_NONE,
                 scancode;
           KeyEvent event;
           Key key;     

           semaphore_down (& keyboardSemaphore);

           kbc_lock ();

           if ((kbc_read_status () & KBC_STATUS_OUTPUT_PENDING) == 0)
           {
              kbc_unlock ();

              continue;
           }

           scancode = kbc_read_data ();

           if (scancode == 0xE0 ||
               scancode == 0xE1)
           { 
              keyboardCurrent -> keyModifiers |= KEY_MODIFIER_EXTENDED;
         
              semaphore_down (& keyboardSemaphore);

              scancode = kbc_read_data ();
           }

           key = key_translation_translate (keyboardCurrent -> keyTranslation,
                                            keyboardCurrent -> keyModifiers,
                                            scancode & KEYBOARD_SCANCODE_KEY);

           if (keyboardCurrent -> keyModifiers & KEY_MODIFIER_EXTENDED)
             keyboardCurrent -> keyModifiers &= ~ KEY_MODIFIER_EXTENDED;

           switch (key)
           {
           case KEY_SHIFT_LEFT:
           case KEY_SHIFT_RIGHT:
                   modifier = KEY_MODIFIER_SHIFT;
 
                   break;
 
           case KEY_ALT_LEFT:
           case KEY_ALT_RIGHT:
                   modifier = KEY_MODIFIER_ALT;
 
                   break;
 
           case KEY_CTRL_LEFT:
           case KEY_CTRL_RIGHT:
                   modifier = KEY_MODIFIER_CTRL;
                    
                   break;
                    
           case KEY_SCROLL_LOCK:
                   modifier = KEY_MODIFIER_SCROLL_LOCK;
                   led      = KEYBOARD_LED_SCROLL_LOCK;
                    
                   break;
 
           case KEY_NUM_LOCK:
                   modifier = KEY_MODIFIER_NUM_LOCK;
                   led      = KEYBOARD_LED_NUM_LOCK;
 
                   break;

           case KEY_CAPS_LOCK:
                   modifier = KEY_MODIFIER_CAPS_LOCK;
                   led      = KEYBOARD_LED_CAPS_LOCK;
          
                   break;
           } 

           if (modifier != KEY_MODIFIER_NONE)
           {
              if (led != KEYBOARD_LED_NONE)
              {
                 if ((scancode & KEYBOARD_SCANCODE_RELEASED) == 0)
                 {
                    keyboardCurrent -> keyModifiers ^= modifier;
                    keyboardCurrent -> leds         ^= led;

                    keyboard_set_leds (keyboardCurrent, keyboardCurrent -> leds);
                 }
              }
              else
              if (scancode & KEYBOARD_SCANCODE_RELEASED)
                keyboardCurrent -> keyModifiers &= ~ modifier;
              else
                keyboardCurrent -> keyModifiers |= modifier;
           }

           event.key   = key;
           event.state = (scancode & KEYBOARD_SCANCODE_RELEASED ? KEY_STATE_RELEASED : KEY_STATE_PRESSED);
                    
           if (keyboardCurrent -> eventFunction != NULL)
             (* keyboardCurrent -> eventFunction) (keyboardCurrent, & event, keyboardCurrent -> eventData);
           else
             keyboard_add_event (keyboardCurrent, & event);

           kbc_unlock ();
        }
}

static Bool
keyboard_initialize (void)
{
   uint8 commandByte;

   kbc_send_command (KBC_COMMAND_ACTIVATE_KEYBOARD);
   
   if (kbc_acknowledge_data (KBC_COMMAND_DISABLE_SCANNING) == FALSE)
     return FALSE;
  
   kbc_send_command (KBC_COMMAND_READ_COMMAND_BYTE);
   commandByte = kbc_receive_data ();

   commandByte |= KBC_COMMAND_BYTE_ENABLE_KEYBOARD_INTERRUPT;
   commandByte |= KBC_COMMAND_BYTE_XT_SCANCODES;
   commandByte &= ~ KBC_COMMAND_BYTE_KEYBOARD_DISABLED;

   kbc_send_command (KBC_COMMAND_WRITE_COMMAND_BYTE);
   kbc_wait_for_write ();
   kbc_write_data (commandByte);

   if (kbc_acknowledge_data (KBC_COMMAND_ENABLE_KEYBOARD) == FALSE)
     return FALSE;
   
   return TRUE;
}

Bool
__keyboard_enter__ (void)
{
        semaphore_create (& keyboardSemaphore, 0);

        interrupt_handler_create (& keyboardHandler,
                                  1,
                                  keyboard_interrupt_handler,
                                  NULL,
                                  0,
                                  "keyboard");

        keyboard_create (& keyboardDefault,
                         NULL,
                         NULL,
                         NULL,
                         NULL,
                         NULL,
                         NULL,
                         NULL);

        if (thread_create (& keyboardThread,
                           THREAD_PRIORITY_MAXIMUM,
                           THREAD_PRIORITY_MAXIMUM,
                           keyboardEventDispatcher,
                           NULL,
                           NULL,
                           0) == FALSE)
          return FALSE;

        if (keyboard_initialize () == FALSE ||
            keyboard_activate (NULL) == FALSE)
        {
           console_print (NULL, "KBC: Failed to initialize keyboard!\n");

           return FALSE;
        }

        console_print (NULL, "KBC: Keyboard initialized.\n");

        return TRUE;
}

void
__keyboard_exit__ (void)
{
        if (keyboardThread != NULL)
          thread_destroy (keyboardThread);

        interrupt_handler_destroy (& keyboardHandler);

        semaphore_destroy (& keyboardSemaphore);
}

Bool
keyboard_create (Keyboard * keyboard,
                 const char * translationName,
                 KeyboardEnterFunction * enterFunction,
                 void * enterData,
                 KeyboardExitFunction * exitFunction,
                 void * exitData,
                 KeyboardEventFunction * eventFunction,
                 void * eventData)
{
        keyboard -> leds           = KEYBOARD_LED_DEFAULT;
        keyboard -> typematicDelay = KEYBOARD_TYPEMATIC_DELAY_DEFAULT;
        keyboard -> typematicRate  = KEYBOARD_TYPEMATIC_RATE_DEFAULT;
        keyboard -> scancodeSet    = KEYBOARD_SCANCODE_SET_DEFAULT;

        keyboard -> enterFunction = enterFunction;
        keyboard -> enterData     = enterData;
        keyboard -> exitFunction  = exitFunction;
        keyboard -> exitData      = exitData;
        keyboard -> eventFunction = eventFunction;
        keyboard -> eventData     = eventData;
       
        condition_create (& keyboard -> eventCondition);
        mutex_create (& keyboard -> eventLock);

        keyboard -> eventsWanted     = 0;
        keyboard -> eventBufferBegin = 0;
        keyboard -> eventBufferEnd   = 0;
        
        keyboard -> keyModifiers     = KEY_MODIFIER_DEFAULT;
       
        if (translationName == NULL)
          keyboard -> keyTranslation = key_translation_default ();
        else
          keyboard -> keyTranslation = key_translation_find (translationName);

        if (keyboard -> keyTranslation == NULL)
          return FALSE;

        return TRUE;
}

void
keyboard_destroy (Keyboard * keyboard)
{
        mutex_destroy (& keyboard -> eventLock);
        condition_destroy (& keyboard -> eventCondition);
        
        if (keyboard == keyboard_current ())
          keyboard_activate (NULL);
}

Keyboard *
keyboard_current (void)
{
        return keyboardCurrent;
}

void
keyboard_set_leds (Keyboard * keyboard, uint8 leds)
{
        kbc_lock ();

        keyboard -> leds = leds;

        if (keyboard == keyboard_current ())
        {
           kbc_acknowledge_data (KBC_COMMAND_SET_KEYBOARD_LEDS);
           kbc_acknowledge_data (leds);
        }

        kbc_unlock ();
}
 
void
keyboard_set_typematic (Keyboard * keyboard, uint8 rate, uint8 delay)
{
        kbc_lock ();

        keyboard -> typematicRate  = rate;
        keyboard -> typematicDelay = delay;

        if (keyboard == keyboard_current ())
        {
           kbc_acknowledge_data (KBC_COMMAND_TYPEMATIC_RATE_PROGRAMMING);
           kbc_acknowledge_data ((delay << 5) | rate);
        }

        kbc_unlock ();
}

void
keyboard_select_scancode_set (Keyboard * keyboard, uint8 scancodeSet)
{
        kbc_lock ();

        keyboard -> scancodeSet = scancodeSet;

        if (keyboard == keyboard_current ())
        {
           kbc_acknowledge_data (KBC_COMMAND_SELECT_SCANCODE_SET);
           kbc_acknowledge_data (scancodeSet);
        }

        kbc_unlock ();
}

Bool
keyboard_activate (Keyboard * keyboard)
{
        if (keyboard == NULL)
          keyboard = & keyboardDefault;

        keyboardCurrent = keyboard;

        return TRUE;

        kbc_lock ();

        interrupt_disable (1);

        if (keyboardCurrent != NULL &&
            keyboardCurrent -> exitFunction != NULL)
          (* keyboardCurrent -> exitFunction) (keyboardCurrent,
                                               keyboardCurrent -> exitData);

        keyboardCurrent = keyboard;

        kbc_acknowledge_data (KBC_COMMAND_SET_KEYBOARD_LEDS);
        kbc_acknowledge_data (keyboard -> leds);
        kbc_acknowledge_data (KBC_COMMAND_TYPEMATIC_RATE_PROGRAMMING);
        kbc_acknowledge_data ((keyboard -> typematicDelay << 5) |
                              keyboard -> typematicRate);
        kbc_acknowledge_data (KBC_COMMAND_SELECT_SCANCODE_SET);
        kbc_acknowledge_data (keyboard -> scancodeSet);
        
        if (keyboard -> enterFunction != NULL)
          (* keyboard -> enterFunction) (keyboard,
                                         keyboard -> enterData);

        interrupt_enable (1);

        kbc_unlock ();

        return TRUE;
}

Bool
keyboard_add_event (Keyboard * keyboard, const KeyEvent * event)
{
        Bool bufferNotFull = TRUE;

        mutex_lock (& keyboard -> eventLock);

        keyboard -> eventBuffer [keyboard -> eventBufferEnd] = * event;

        ++ keyboard -> eventBufferEnd;

        if (keyboard -> eventBufferEnd >= KEYBOARD_EVENT_BUFFER_SIZE)
          keyboard -> eventBufferEnd = 0;

        if (keyboard -> eventBufferEnd == keyboard -> eventBufferBegin)
        {
           bufferNotFull = FALSE;

           ++ keyboard -> eventBufferBegin;

           if (keyboard -> eventBufferBegin >= KEYBOARD_EVENT_BUFFER_SIZE)
             keyboard -> eventBufferBegin = 0;
        }

        if (keyboard -> eventsWanted > 0)
        {
           -- keyboard -> eventsWanted;

           if (keyboard -> eventsWanted <= 0)
             condition_signal (& keyboard -> eventCondition);
        }

        mutex_unlock (& keyboardCurrent -> eventLock);

        return bufferNotFull;
}

Size
keyboard_peek_events (Keyboard * keyboard, KeyEvent * events, Size eventsWanted)
{
        Size eventsPending;

        if (keyboard == NULL)
          keyboard = & keyboardDefault;

        mutex_lock (& keyboard -> eventLock);

        if (keyboard -> eventBufferEnd >= keyboard -> eventBufferBegin)
        {
           eventsPending = keyboard -> eventBufferEnd - keyboard -> eventBufferBegin;
        
           if (eventsPending > eventsWanted)
             eventsPending = eventsWanted;

           memory_copy (events, 
                        & keyboard -> eventBuffer [keyboard -> eventBufferBegin], 
                        eventsPending * sizeof (KeyEvent));
        }
        else
        {
           eventsPending = KEYBOARD_EVENT_BUFFER_SIZE - keyboard -> eventBufferBegin;

           if (eventsPending > eventsWanted)
             eventsPending = eventsWanted;

           eventsWanted -= eventsPending;

           memory_copy (events, 
                        & keyboard -> eventBuffer [keyboard -> eventBufferBegin], 
                        eventsPending * sizeof (KeyEvent));

           if (eventsWanted > 0)
           {
              if (keyboard -> eventBufferEnd < eventsWanted)
                eventsWanted = keyboard -> eventBufferEnd;

              memory_copy (& events [eventsPending], 
                           keyboard -> eventBuffer,
                           eventsWanted * sizeof (KeyEvent));
           }

           eventsPending += eventsWanted;
        }

        keyboard -> eventBufferBegin += eventsPending;

        if (keyboard -> eventBufferBegin >= KEYBOARD_EVENT_BUFFER_SIZE)
          keyboard -> eventBufferBegin -= KEYBOARD_EVENT_BUFFER_SIZE;

        mutex_unlock (& keyboard -> eventLock);

        return eventsPending;
}

Size
keyboard_read_events (Keyboard * keyboard, KeyEvent * events, Size eventsWanted)
{
        KeyEvent * eventCurrent = events;

        if (keyboard == NULL)
          keyboard = & keyboardDefault;

        mutex_lock (& keyboard -> eventLock);

        while (keyboard -> eventsWanted <= 0)
        {
            Size eventsRead = keyboard_peek_events (keyboard, eventCurrent, eventsWanted);

            eventsWanted -= eventsRead;

            eventCurrent += eventsRead;
            
            if (eventsWanted <= 0)
              break;

            if (eventsWanted > KEYBOARD_EVENT_BUFFER_SIZE - 1)
              keyboard -> eventsWanted = KEYBOARD_EVENT_BUFFER_SIZE - 1;
            else
              keyboard -> eventsWanted = eventsWanted;

            condition_wait (& keyboard -> eventCondition, & keyboard -> eventLock);
        } 

        mutex_unlock (& keyboard -> eventLock);

        return (Size) (eventCurrent - events);
}

                           
               

