#ifndef __KBC_KEYBOARD_H__
#define __KBC_KEYBOARD_H__

#include <utility/types.h>
#include <scheduler/condition.h>
#include <scheduler/mutex.h>
#include <kbc/key_translation.h>
#include <kbc/key_event.h>

struct Keyboard;

typedef enum
{
        KEYBOARD_LED_NONE        = 0,
        KEYBOARD_LED_DEFAULT     = 0,

        KEYBOARD_LED_SCROLL_LOCK = (1 << 0),
        KEYBOARD_LED_NUM_LOCK    = (1 << 1),
        KEYBOARD_LED_CAPS_LOCK   = (1 << 2)
} KeyboardLED;

typedef enum
{
        KEYBOARD_TYPEMATIC_RATE_DEFAULT = 0,

        KEYBOARD_TYPEMATIC_RATE_MINIMUM = 0,
        KEYBOARD_TYPEMATIC_RATE_MAXIMUM = 31
} KeyboardTypematicRate;

typedef enum
{
        KEYBOARD_TYPEMATIC_DELAY_DEFAULT = 0,

        KEYBOARD_TYPEMATIC_DELAY_MINIMUM = 0,
        KEYBOARD_TYPEMATIC_DELAY_MAXIMUM = 3
} KeyboardTypematicDelay;

typedef enum
{
        KEYBOARD_SCANCODE_SET_DEFAULT = 2,
        
        KEYBOARD_SCANCODE_SET_1       = 1,
        KEYBOARD_SCANCODE_SET_2       = 2,
        KEYBOARD_SCANCODE_SET_3       = 3
} KeyboardScancodeSet;

typedef enum
{
        KEYBOARD_SCANCODE_KEY      = 0x7F,
        KEYBOARD_SCANCODE_PRESSED  = 0x00,
        KEYBOARD_SCANCODE_RELEASED = 0x80
} KeyboardScancode; 

typedef void KeyboardEnterFunction (struct Keyboard *, void *);
typedef void KeyboardExitFunction (struct Keyboard *, void *);
typedef void KeyboardEventFunction (struct Keyboard *, const KeyEvent *, void *);

enum
{
        KEYBOARD_EVENT_BUFFER_SIZE = 1024
};

typedef struct Keyboard
{
        uint8                   leds;
        uint8                   typematicRate;
        uint8                   typematicDelay;
        uint8                   scancodeSet;
        KeyboardEnterFunction * enterFunction;
        void *                  enterData;
        KeyboardExitFunction *  exitFunction;
        void *                  exitData;
        KeyboardEventFunction * eventFunction;
        void *                  eventData;
        Condition               eventCondition;
        Mutex                   eventLock;
        Size                    eventsWanted;
        KeyTranslation *        keyTranslation;
        KeyModifiers            keyModifiers;
        Size                    eventBufferBegin;
        Size                    eventBufferEnd;
        KeyEvent                eventBuffer [KEYBOARD_EVENT_BUFFER_SIZE];
} Keyboard;

extern Bool keyboard_create (Keyboard *, 
                             const char *,
                             KeyboardEnterFunction *,
                             void *,
                             KeyboardExitFunction *,
                             void *,
                             KeyboardEventFunction *,
                             void *);
extern void keyboard_destroy (Keyboard *);

extern Bool keyboard_activate            (Keyboard *);
extern Keyboard * keyboard_current       (void);

extern void keyboard_set_leds            (Keyboard *, uint8);
extern void keyboard_set_typematic       (Keyboard *, uint8, uint8);
extern void keyboard_select_scancode_set (Keyboard *, uint8);

extern Bool keyboard_add_event           (Keyboard *, const KeyEvent *);
extern Size keyboard_peek_events         (Keyboard *, KeyEvent *, Size);
extern Size keyboard_read_events         (Keyboard *, KeyEvent *, Size);

#endif /* __KBC_KEYBOARD_H__ */

