#include <utility/memory.h>
#include <memory/allocator.h>

Bool
memory_allocator_create (MemoryAllocator * allocator)
{
   Size alignment;

   for (alignment = 0;
        alignment < SLAB_ALIGNMENT;
	++ alignment)
     cache_create (& allocator -> caches [alignment], (1 << alignment));
  
   return TRUE;
}

void
memory_allocator_destroy (MemoryAllocator * allocator)
{
   Size alignment;

   for (alignment = 0;
        alignment < SLAB_ALIGNMENT;
	++ alignment)
     cache_destroy (& allocator -> caches [alignment]);
}

void *
memory_allocator_allocate_within (MemoryAllocator * allocator, void * begin, void * end, Size size)
{
   void * result = NULL;
   Size alignment = memory_alignment (SLAB_ALIGNMENT, sizeof (Size) * 8, size);

   result = (void *) (SLAB_SIZE * page_allocate_within (alignment - SLAB_ALIGNMENT, 
                                                        (PageAddress) begin / SLAB_SIZE,
                                                        (PageAddress) end / SLAB_SIZE,
                                                        allocator));
   if (result != NULL)
     return result;

   cache_reap (~0);

   return (void *) (SLAB_SIZE * page_allocate_within (alignment - SLAB_ALIGNMENT,
                                                      (PageAddress) begin / SLAB_SIZE,
                                                      (PageAddress) end / SLAB_SIZE,
                                                      allocator));
}
   
void *
memory_allocator_allocate (MemoryAllocator * allocator, Size size)
{
   void * result = NULL;
   Size alignment = memory_alignment (0, sizeof (Size) * 8, size);

   if (alignment < SLAB_ALIGNMENT)
     result = cache_allocate (& allocator -> caches [alignment]);
   else
     result = (void *) (SLAB_SIZE * page_allocate (alignment - SLAB_ALIGNMENT, PAGE_ALLOCATOR_ATTRIBUTES_ANY, allocator));

   if (result != NULL)
     return result;
   
   cache_reap (~0);
   
   if (alignment < SLAB_SIZE)
     return cache_allocate (& allocator -> caches [alignment]);
  
   return (void *) (SLAB_SIZE * page_allocate (alignment - SLAB_ALIGNMENT, PAGE_ALLOCATOR_ATTRIBUTES_ANY, allocator));
}

void
memory_allocator_free (MemoryAllocator * allocator, void * memory)
{
   PageAddress address = (PageAddress) memory / SLAB_SIZE;
   void *      owner   = page_get_owner (address);

   if (owner == NULL)
     return;

   if (owner == allocator)
     page_free (address);
   else
   if ((Cache *) owner >= allocator -> caches &&
       (Cache *) owner < & allocator -> caches [SLAB_ALIGNMENT])
     cache_free ((Cache *) owner, memory);
}

void *
memory_allocator_resize (MemoryAllocator * allocator, void * memory, Size size)
{
   void * result;
   Size   oldSize;

   if (size <= 0)
   {
      memory_allocator_free (allocator, memory);

      return NULL;
   }

   result = memory_allocator_allocate (allocator, size);
   if (result == NULL ||
       memory == NULL)
     return result;

   oldSize = memory_allocator_size (allocator, memory);
   if (oldSize <= 0)
     return result;

   if (size < oldSize)
     oldSize = size;

   memory_copy (result, memory, oldSize);

   memory_allocator_free (allocator, memory);

   return result;
}

void
memory_allocator_claim (MemoryAllocator * allocator, void * memory, Size size)
{
   PageAddress pageAddress = (PageAddress) memory >> SLAB_ALIGNMENT;
   Size        pageCount   = (size + SLAB_SIZE - 1) >> SLAB_ALIGNMENT;

   while (pageCount > 0)
   {
       page_allocate_within (0, pageAddress, pageAddress, allocator);

       ++ pageAddress;
       -- pageCount;
   }
}

void
memory_allocator_release (MemoryAllocator * allocator, void * memory, Size size)
{
   PageAddress pageAddress = (PageAddress) memory >> SLAB_ALIGNMENT;
   Size        pageCount   = (size + SLAB_SIZE - 1) >> SLAB_ALIGNMENT;

   while (pageCount > 0)
   {
       if (page_get_owner (pageAddress) == allocator)
         page_free (pageAddress);

       ++ pageAddress;
       -- pageCount;
   }
}

Size
memory_allocator_size (MemoryAllocator * allocator, void * memory)
{
   PageAddress address = (PageAddress) memory / SLAB_SIZE;
   void *      owner   = page_get_owner (address);

   if (owner == NULL)
     return 0;

   if (owner == allocator)
     return page_get_size (address) * SLAB_SIZE;
   
   return ((Cache *) owner) -> objectSize;
}

