#include <utility/memory.h>
#include <persistence/mapping.h>

static List mappingRanges;
static List dirtyMappings;
static List cleanMappings;

Size
mapping_range_size (Address rangeBegin,
                    Address rangeEnd)
{
        return sizeof (MappingRange) +
               sizeof (Mapping) * (rangeEnd - rangeBegin) / I386_PAGE_SIZE;
}

Bool
mapping_range_create (MappingRange * range,
                      Address rangeBegin,
                      Address rangeEnd)
{
        range -> rangeBegin = rangeBegin;
        range -> rangeEnd   = rangeEnd;
        range -> mappings   = (Mapping *) & range [1];
        
        memory_zero (range -> mappings,
                     sizeof (Mapping) * (rangeEnd - rangeBegin) / I386_PAGE_SIZE);

        return TRUE;
}

void
mapping_range_insert (MappingRange * newRange)
{
        ListIterator position;

        for (position = list_begin (& mappingRanges);
             position != list_end (& mappingRanges);
             position = list_next (position))
        {
            MappingRange * range = (MappingRange *) position;

            if (newRange -> rangeBegin < range -> rangeBegin)
              break;
        }

        list_insert (position, newRange);
}

MappingRange *
mapping_range_find (const Mapping * mapping)
{
        ListIterator position;
       
        for (position = list_begin (& mappingRanges);
             position != list_end (& mappingRanges);
             position = list_next (position))
        {
            MappingRange * range = (MappingRange *) position;

            if (mapping < range -> mappings)
              break;
        }

        position = list_previous (position); 
        
        return (MappingRange *) position;
}

Mapping *
mapping_allocate (Address physicalAddress, Size mappings, Address virtualAddress, FlagSet flags)
{
       Mapping * mapping, 
               * mappingCurrent;
       i386PageEntry * pageEntry;

       pageEntry  = (i386PageEntry *) ((I386_PAGE_DIRECTORY_ENTRIES - 1) << I386_PAGE_DIRECTORY_SHIFT);
       pageEntry += virtualAddress / I386_PAGE_SIZE;

       mapping = mapping_find (physicalAddress);

       for (mappingCurrent = mapping;
            mappings > 0;
            ++ mappingCurrent,
               -- mappings,
               ++ pageEntry,
               physicalAddress += I386_PAGE_SIZE)
       {
          * pageEntry = physicalAddress | flags | MAPPING_FLAG_PRESENT;

          mappingCurrent -> pageEntry = pageEntry;

          list_insert_front (& cleanMappings, mappingCurrent);
       }

       return mapping;
}

void
mapping_free (Address physicalAddress, Size mappings)
{
        Mapping * mapping = mapping_find (physicalAddress);

        while (mappings > 0)
        {
            list_remove (& mapping -> mappingList);

            * mapping -> pageEntry = ~ MAPPING_FLAG_PRESENT;

            mapping -> pageEntry = NULL;

            ++ mapping;
            -- mappings;
        }
}
        
Mapping *
mapping_find (Address address)
{
        ListIterator position;

        for (position = list_begin (& mappingRanges);
             position != list_end (& mappingRanges);
             position = list_next (position))
        {
            MappingRange * range = (MappingRange *) position;

            if (address >= range -> rangeEnd)
              break;

            if (address >= range -> rangeBegin)
              return & range -> mappings [address - range -> rangeBegin];
        }

        return NULL;
}

Address
mapping_virtual_address (const Mapping * mapping)
{
        Address address = (Address) page -> pageEntry;

        address -= (I386_PAGE_DIRECTORY_ENTRIES - 1) << I386_PAGE_DIRECTORY_SHIFT;
        address *= I386_PAGE_SIZE / sizeof (i386PageEntry);

        return address;
}

Address
mapping_physical_address (const Mapping * mapping)
{
        MappingRange * range = mapping_range_find (mapping);

        return range -> rangeBegin + (mapping - range -> mappings) * I386_PAGE_SIZE;
}

