#include <scheduler/critical.h>
#include <scheduler/thread.h>
#include <scheduler/mutex.h>

Bool
mutex_create (Mutex * mutex)
{
   spinlock_create (& mutex -> lock);

   mutex -> owner = NULL;
   mutex -> depth = 0;
   
   list_clear (& mutex -> waitingThreads);
 
   return TRUE;
}

void
mutex_destroy (Mutex * mutex)
{
   spinlock_lock (& mutex -> lock);
   
   while (list_empty (& mutex -> waitingThreads) == FALSE)
   {
       Thread * waiting = (Thread *) list_remove_front (& mutex -> waitingThreads);

       thread_start (waiting);
   }

   spinlock_unlock (& mutex -> lock);

   spinlock_destroy (& mutex -> lock);
}

Size
mutex_lock (Mutex * mutex)
{
   Thread * current;

   spinlock_lock (& mutex -> lock);

   current = thread_current ();

   if (mutex -> owner == NULL ||
       mutex -> owner == current)
   {
       mutex -> owner = current;
       ++ mutex -> depth;

       spinlock_unlock (& mutex -> lock);

       return mutex -> depth;
   }

   thread_block (current, THREAD_STATE_WAITING_ON_MUTEX, & mutex -> lock, & mutex -> waitingThreads);
  
   mutex -> owner = current;
   mutex -> depth = 1;

   spinlock_unlock (& mutex -> lock);

   return mutex -> depth;
}

typedef struct
{
    Thread * thread;
    Mutex *  mutex;
} TimedLockContext;

static void
timedLockTimeoutHandler (Timer * timer, void * data)
{
   TimedLockContext * context = (TimedLockContext *) data;

   spinlock_lock (& context -> mutex -> lock);

   list_remove (& context -> thread -> threadList);

   thread_start (context -> thread);

   spinlock_unlock (& context -> mutex -> lock);
}

Size
mutex_timed_lock (Mutex * mutex, TimeOffset * timeout)
{
   Size depth;
   Thread * current;
   Timer timer;
   TimedLockContext context;

   spinlock_lock (& mutex -> lock);

   current = thread_current ();

   context.thread = current;
   context.mutex  = mutex;
  
   if (mutex -> owner == NULL ||
       mutex -> owner == current)
   {
       mutex -> owner = current;
       ++ mutex -> depth;

       spinlock_unlock (& mutex -> lock);

       return mutex -> depth;
   }

   timer_create (& timer, timeout, timedLockTimeoutHandler, & context, FALSE);

   critical_enter ();

   timer_start (& timer);

   thread_block (current, THREAD_STATE_WAITING_ON_MUTEX, & mutex -> lock, & mutex -> waitingThreads);
   
   timer_stop (& timer, timeout);

   if (mutex -> owner != NULL)
   {  
      mutex -> owner = current;
      mutex -> depth = 1;

      depth = 1;
   }
   else
      depth = 0;

   spinlock_unlock (& mutex -> lock);

   timer_destroy (& timer);

   return depth;
}
        
Size
mutex_try_lock (Mutex * mutex)
{
   spinlock_lock (& mutex -> lock);

   if (mutex -> owner == NULL ||
       mutex -> owner == thread_current ())
   {
       mutex -> owner = thread_current ();
       mutex -> depth = 1;

       spinlock_unlock (& mutex -> lock);

       return mutex -> depth;
   }

   spinlock_unlock (& mutex -> lock);

   return 0;
}

Size
mutex_unlock (Mutex * mutex)
{
   Size newDepth;

   spinlock_lock (& mutex -> lock);

   newDepth = -- mutex -> depth;
   
   if (newDepth <= 0)
   {
      mutex -> owner = NULL;

      if (list_empty (& mutex -> waitingThreads) == FALSE)
      {
	 Thread * waiting = (Thread *) list_remove_front (& mutex -> waitingThreads);

	 thread_start (waiting);
      }
   }

   spinlock_unlock (& mutex -> lock);

   return newDepth;
}

