#include <scheduler/critical.h>
#include <scheduler/semaphore.h>

Bool
semaphore_create (Semaphore * semaphore, Size count)
{
   spinlock_create (& semaphore -> lock);

   semaphore -> count = 0;
   
   list_clear (& semaphore -> waitingThreads);
 
   return TRUE;
}

void
semaphore_destroy (Semaphore * semaphore)
{
   spinlock_lock (& semaphore -> lock);
   
   while (list_empty (& semaphore -> waitingThreads) == FALSE)
   {
       Thread * waiting = (Thread *) list_remove_front (& semaphore -> waitingThreads);

       thread_start (waiting);
   }

   spinlock_unlock (& semaphore -> lock);

   spinlock_destroy (& semaphore -> lock);
}

Size
semaphore_down (Semaphore * semaphore)
{
   spinlock_lock (& semaphore -> lock);

   if (semaphore -> count > 0)
   {
       -- semaphore -> count;

       spinlock_unlock (& semaphore -> lock);

       return semaphore -> count;
   }

   critical_enter ();

   if (semaphore -> count <= 0)
     thread_block (NULL, THREAD_STATE_WAITING_ON_SEMAPHORE, & semaphore -> lock, & semaphore -> waitingThreads);

   -- semaphore -> count;

   spinlock_unlock (& semaphore -> lock);

   critical_exit ();

   return semaphore -> count;
}

typedef struct
{
   Thread *    thread;
   Semaphore * semaphore;
} TimedDownContext;

static void
timedDownTimeoutHandler (Timer * timer, void * data)
{
   TimedDownContext * context = (TimedDownContext *) data;

   spinlock_lock (& context -> semaphore -> lock);

   list_remove (& context -> thread -> threadList);

   thread_start (context -> thread);

   spinlock_unlock (& context -> semaphore -> lock);
}

Size
semaphore_timed_down (Semaphore * semaphore, TimeOffset * timeout)
{
   Size count;
   Timer timer;
   TimedDownContext context;

   context.thread    = thread_current ();
   context.semaphore = semaphore;

   spinlock_lock (& semaphore -> lock);

   if (semaphore -> count > 0)
   {
       -- semaphore -> count;

       spinlock_unlock (& semaphore -> lock);

       return semaphore -> count;
   }

   timer_create (& timer, timeout, timedDownTimeoutHandler, & context, FALSE);

   critical_enter ();

   timer_start (& timer);

   if (semaphore -> count <= 0)
     thread_block (NULL, THREAD_STATE_WAITING_ON_SEMAPHORE, & semaphore -> lock, & semaphore -> waitingThreads);

   if (semaphore -> count > 0)
     count = -- semaphore -> count;
   else
     count = ~0;

   timer_stop (& timer, timeout);

   spinlock_unlock (& semaphore -> lock);

   critical_exit ();

   timer_destroy (& timer);

   return count;
}

Size
semaphore_try_down (Semaphore * semaphore)
{
   spinlock_lock (& semaphore -> lock);

   if (semaphore -> count > 0)
   {
       -- semaphore -> count;

       spinlock_unlock (& semaphore -> lock);

       return semaphore -> count;
   }

   spinlock_unlock (& semaphore -> lock);

   return ~0;
}

Size
semaphore_up (Semaphore * semaphore)
{
   Size newCount;

   critical_enter ();
   
   newCount = ++ semaphore -> count;

   while (list_empty (& semaphore -> waitingThreads) == FALSE &&
	  newCount > 0)
   {
       Thread * waiting = (Thread *) list_remove_front (& semaphore -> waitingThreads);

       thread_start (waiting);

       -- newCount;
   }

   critical_exit ();

   return semaphore -> count;
}

