#include <memory/memory.h>
#include <utility/memory.h>
#include <utility/string.h>
#include <utility/hash_table.h>

Bool
hash_table_create (HashTable * table, Size nodeWidth)
{
   if (nodeWidth < 1)
     nodeWidth = 1;

   table -> nodeWidth   = nodeWidth;
   table -> nodeCount   = 0;
   table -> bucketCount = 0;
   table -> buckets     = NULL;

   return hash_table_grow (table, HASH_TABLE_EMPTY_THRESHOLD * HASH_TABLE_GROWTH_SCALE + 
                                  HASH_TABLE_GROWTH_OFFSET);
}

void
hash_table_destroy (HashTable * table)
{
   hash_table_clear (table);
}

void
hash_table_clear (HashTable * table)
{
    List * currentBucket;

    if (table -> bucketCount <= 0)
      return;

    for (currentBucket = table -> buckets;
         currentBucket < & table -> buckets [table -> bucketCount];
         ++ currentBucket)
    {
        while (list_empty (currentBucket) == FALSE)
           memory_free (list_remove_front (currentBucket));
    }

    memory_free (table -> buckets);

    table -> nodeCount   = 0;
    table -> bucketCount = 0;
    table -> buckets     = NULL;
}

void **
hash_table_find (HashTable * table, const char * key)
{
   Size   hash   = string_hash (key) % table -> bucketCount;
   List * bucket = & table -> buckets [hash];
   ListIterator position;

   for (position = list_begin (bucket);
        position != list_end (bucket);
        position = list_next (position))
   {
       HashTableNode * node = (HashTableNode *) position;

       if (string_equal (key, node -> key) == TRUE)
         return node -> values;
   }
   
   return NULL;
}

Bool
hash_table_remove (HashTable * table, const char * key)
{
   Size   hash   = string_hash (key) % table -> bucketCount;
   List * bucket = & table -> buckets [hash];
   ListIterator position;

   for (position = list_begin (bucket);
        position != list_end (bucket);
        position = list_next (position))
   {
       HashTableNode * node = (HashTableNode *) position;

       if (string_equal (key, node -> key) == TRUE)
       {
          -- table -> nodeCount;

          list_remove (position);

          memory_free (node);

          return TRUE;
       }
   }

   return FALSE;
}

void **
hash_table_insert (HashTable * table, const char * key)
{
    Size hash;
    List * bucket;
    ListIterator position;
    HashTableNode * node;

    if (table -> bucketCount - table -> nodeCount <= HASH_TABLE_EMPTY_THRESHOLD)
      hash_table_grow (table, table -> nodeCount * HASH_TABLE_GROWTH_SCALE +
                              HASH_TABLE_GROWTH_OFFSET);
    
    hash   = string_hash (key) % table -> bucketCount;
    bucket = & table -> buckets [hash];

    for (position = list_begin (bucket);
         position != list_end (bucket);
         position = list_next (position))
    {
        node = (HashTableNode *) position;

        if (string_equal (key, node -> key) == TRUE)
	  return node -> values;
    }
    
    node = (HashTableNode *) memory_allocate (sizeof (HashTableNode) + (table -> nodeWidth - 1) * sizeof (void *));
    if (node == NULL)
      return NULL;
   
    node -> key = key;

    memory_zero (node -> values, table -> nodeWidth * sizeof (void *));
    
    list_insert (position, node);

    ++ table -> nodeCount;

    return node -> values;
}

Bool
hash_table_grow (HashTable * table, Size newSize)
{
    List * newBuckets,
         * currentBucket;

    if (newSize <= table -> bucketCount)
      return TRUE;

    if (newSize < HASH_TABLE_EMPTY_THRESHOLD * HASH_TABLE_GROWTH_SCALE + HASH_TABLE_GROWTH_OFFSET)
      newSize = HASH_TABLE_EMPTY_THRESHOLD * HASH_TABLE_GROWTH_SCALE + HASH_TABLE_GROWTH_OFFSET;

    newBuckets = (List *) memory_allocate (newSize * sizeof (List));
    if (newBuckets == NULL)
      return FALSE;

    for (currentBucket = newBuckets;
         currentBucket < & newBuckets [newSize];
         ++ currentBucket)
      list_clear (currentBucket);

    if (table -> bucketCount > 0)
    {
       for (currentBucket = table -> buckets;
            currentBucket < & table -> buckets [table -> bucketCount];
            ++ currentBucket)
       {
           ListIterator position = list_begin (currentBucket);

           while (position != list_end (currentBucket))
           {
               HashTableNode * node = (HashTableNode *) position;
               Size            hash = string_hash (node -> key) % newSize;

               position = list_next (position);

               list_move_to_back (& newBuckets [hash], & node -> bucketList);
           } 
       }
 
       memory_free (table -> buckets);
    }

    table -> bucketCount = newSize;
    table -> buckets     = newBuckets;

    return TRUE;
}

void
hash_table_for_each (HashTable * table, Size valueIndex, void (* forEachFunction) (void *))
{
    List * currentBucket;

    for (currentBucket = table -> buckets;
         currentBucket < & table -> buckets [table -> bucketCount];
         ++ currentBucket)
    {
        ListIterator position = list_begin (currentBucket);

        while (position != list_end (currentBucket))
        {
            HashTableNode * node = (HashTableNode *) position;

            (* forEachFunction) (node -> values [valueIndex]);

            position = list_next (position);
        }
    }
}

