#include <memory/memory.h>
#include <utility/memory.h>

Size
memory_alignment (Size alignment, Size maximum, Size size)
{
   Size closest = (1 << alignment);

   while (closest < size &&
	  alignment < maximum)
   {
      closest <<= 1;
      ++ alignment;
   }

   return alignment;
}

void *
memory_zero (void * memory, Size count)
{
   uint8 * currentMemory  = (uint8 *) memory;
   Size currentCount      = count / 4;

   if (currentCount > 0)
   do
   {
      * (uint32 *) currentMemory = 0;

      currentMemory += 4;
      -- currentCount;
   } while (currentCount > 0);

   switch (count % 4)
   {
   case 3:
      * currentMemory ++ = 0;
   
   case 2:
      * currentMemory ++ = 0;
   
   case 1:
      * currentMemory    = 0;
   }
   
   return memory;
}

void *
memory_set (void * memory, int value, Size count)
{
   uint8 * currentMemory  = (uint8 *) memory;
   Size currentCount      = count / 4;

   if (currentCount > 0)
   {
      uint32 chunk = (uint8) value |
	             ((uint8) value << 8) |
		     ((uint8) value << 16) |
		     ((uint8) value << 24);

      do
      {
         * (uint32 *) currentMemory = chunk;

         currentMemory += 4;
         -- currentCount;
      } while (currentCount > 0);
   }
   
   switch (count % 4)
   {
   case 3:
      * currentMemory ++ = (uint8) value;

   case 2:
      * currentMemory ++ = (uint8) value;

   case 1:
      * currentMemory    = (uint8) value;
   }

   return memory;
}

void *
memory_copy (void * dst, const void * src, Size count)
{
   uint8 * currentDst       = (uint8 *) dst;
   const uint8 * currentSrc = (const uint8 *) src;
   Size currentCount        = count / 4;

   if (currentCount > 0)
   do
   {
      * (uint32 *) currentDst = * (const uint32 *) currentSrc;

      currentDst += 4;
      currentSrc += 4;
      -- currentCount;
   } while (currentCount > 0);

   switch (count % 4)
   {
   case 3:
      * currentDst ++ = * currentSrc ++;

   case 2:
      * currentDst ++ = * currentSrc ++;

   case 1:
      * currentDst    = * currentSrc;
   }

   return dst;
}

void *
memory_duplicate (const void * src, Size count)
{
   void * dst = memory_allocate (count);

   if (dst == NULL)
     return NULL;

   memory_copy (dst, src, count);

   return dst;
}

void *
memory_move (void * dst, const void * src, Size count)
{
   uint8 * currentDst;
   const uint8 * currentSrc;
   Size currentCount;

   if (dst < src)
     return memory_copy (dst, src, count);

   currentDst   = (uint8 *) dst + count;
   currentSrc   = (const uint8 *) src + count;
   currentCount = count / 4;

   switch (count % 4)
   {
   case 3:
      * -- currentDst = * -- currentSrc;
   
   case 2:
      * -- currentDst = * -- currentSrc;
      
   case 1:
      * -- currentDst = * -- currentSrc;
   }

   if (currentCount > 0)
   do
   {
      currentDst -= 4;
      currentSrc -= 4;
      -- currentCount;
      
      * (uint32 *) currentDst = * (const uint32 *) currentSrc;
   } while (currentCount > 0);

   return dst;
}

void *
memory_reverse (void * dst, const void * src, Size count)
{
   Size currentCount        = count;
   uint8 * currentDst       = (uint8 *) dst;
   const uint8 * currentSrc = (const uint8 *) src + count - 1;

   if (currentCount > 0)
   do
   {
       * currentDst = * currentSrc;
       
       ++ currentDst;
       -- currentSrc;
       -- currentCount;
   } while (currentCount > 0);

   return dst;
}

Compare
memory_compare (const void * dst, const void * src, Size size)
{
   uint8 * currentDst       = (uint8 *) dst;
   const uint8 * currentSrc = (const uint8 *) src;
   Size currentCount        = size / 4;

   while (currentCount > 0 &&
          * (uint32 *) currentDst == * (const uint32 *) currentSrc)
   {
      currentDst += 4;
      currentSrc += 4;
      -- currentCount;
   }

   if (currentCount > 0)
     size  = 4;
   else
     size %= 4;

   -- currentDst;
   -- currentSrc;

   switch (size)
   {
   case 4:
      if (* ++ currentDst != * ++ currentSrc)
        break;
   
   case 3:
      if (* ++ currentDst != * ++ currentSrc)
        break;

   case 2:
      if (* ++ currentDst != * ++ currentSrc)
        break;

   case 1:
      if (* ++ currentDst != * ++ currentSrc)
        break;

   default:
      return COMPARE_EQUAL;
   }

   if (* currentDst < * currentSrc)
     return COMPARE_LESS;

   return COMPARE_GREATER;
}

void *
memory_find (const void * dst, int value, Size size)
{
   const uint8 * currentDst = (uint8 *) dst;

   while (size > 0 &&
          * currentDst != (uint8) value)
   {
       ++ currentDst;
       -- size;
   }

   if (size > 0)
     return (void *) currentDst;

   return NULL;
}
               
