#include <memory/memory.h>
#include <utility/memory.h>
#include <utility/string.h>

char *
string_copy (char * buffer, Size size, const char * source)
{
    char * current = buffer;
    
    while (* source != '\0' &&
	   size > 1)
    {
        * current = * source;
	
	++ current;
	++ source;
	-- size;
    }

    * current = '\0';
    
    return buffer;
}

char *
string_duplicate (const char * string)
{
    Size length      = string_length (string) + 1;
    char * newString = (char *) memory_allocate (length);

    if (newString == NULL)
      return NULL;

    memory_copy (newString, string, length);

    return newString;
}
    
Size
string_hash (const char * string)
{
    Size character, hash, bit;

    bit  = 0;
    hash = 0;

    for (character = * string;
	 character != '\0';
	 character = * ++ string)
    {
	hash += (character << bit);
	bit   = (bit + 1) % (8 * sizeof (Size));
    }
    
    return hash;
}

Size
string_length (const char * string)
{
    const char * position = string;

    while (* position != '\0')
       ++ position;
    
    return (position - string);
}

char *
string_find (const char * dst, const char * src)
{
    const char * position;

    for (;;)
    {
       while (* dst != * src &&
              * dst != '\0')
          ++ dst;

       if (* dst == '\0')
         return NULL;
    
       position = src;
       do 
       {
           ++ position;
	   ++ dst;
       }
       while (* position == * dst &&
	      * position != '\0');
	   
       if (* position == '\0')
         return (char *) dst - (position - src);
    }
}
    
	     
Bool
string_equal (const char * dst, const char * src)
{
    while (* dst == * src &&
	   * dst != '\0')
    {
       ++ dst;
       ++ src;
    }
    
    return (* dst == * src);
}

Compare
string_compare (const char * dst, const char * src)
{
    while (* dst == * src &&
	   * dst != '\0')
    {
       ++ dst;
       ++ src;
    }

    if (* dst < * src)
      return COMPARE_LESS;

    if (* dst > * src)
      return COMPARE_GREATER;

    return COMPARE_EQUAL;
}

Size
string_format (char * buffer, Size size, const char * format, ...)
{
    Size length;
    Arguments arguments;
    
    ARGUMENTS_BEGIN (arguments, format);

    length = string_format_arguments (buffer, size, format, arguments);

    ARGUMENTS_END (arguments);

    return length;
}

Size
string_format_arguments (char * buffer, Size size, const char * format, Arguments arguments)
{
    char * current = buffer;
    Size precision;

    while (size > 1 &&
	   * format != '\0')
    {
        if (* format != '%')
          * current = * format;
	else
	{
	   ++ format;
	  
           precision = size;
           if (* format == '.')
           {
              precision = string_to_integer (& format [1]) + 1;
              if (precision > size)
                precision = size;

              do ++ format;
              while (* format >= '0' && * format <= '9');
           }

	   switch (* format)
	   {
           case '%':
              * current = '%';
              
              break;
              
           case 'c':
	      * current = ARGUMENTS_NEXT (arguments, char);

	      break;

           case 'n':
              {
                  int * length = ARGUMENTS_NEXT (arguments, int *);

                  if (length != NULL)
                    * length = (int) (current - buffer);

                  continue;
              }

           default:
              switch (* format)
              {
              case 's': 
                 {
                    const char * string = ARGUMENTS_NEXT (arguments, const char *);

                    if (string != NULL)
		      string_copy (current, precision, string);

                    break;
                 }
              
              case 'b':
                 string_from_unsigned (current, precision, ARGUMENTS_NEXT (arguments, unsigned), 2);

                 break;

              case 'u':
		 string_from_unsigned (current, precision, ARGUMENTS_NEXT (arguments, unsigned), 10);

                 break;

              case 'x':
              case 'p':
		 string_from_unsigned (current, precision, ARGUMENTS_NEXT (arguments, unsigned), 16);

                 break;

              case 'd':
              case 'i':
		 string_from_integer (current, precision, ARGUMENTS_NEXT (arguments, int));

                 break;

              default:
                 continue;
              }

	      { 
		  Size length = string_length (current);
		  
	          current += length - 1;
		  size    -= length - 1;
              }

	      break;
           }
	}

	++ current;
	++ format;
	-- size;
    }

    * current = '\0';

    return (Size) (current - buffer);
}
     
char *
string_from_integer (char * string, Size size, int integer)
{
    char buffer [32];
    Bool negative  = FALSE;
    char * current = & buffer [sizeof (buffer)];

    if (integer < 0)
      negative = TRUE;
    
    do
    {
       int remainder = integer % 10;
       if (remainder < 0)
         remainder = - remainder;
       
       integer /= 10;

       -- current;
       * current = (remainder + '0');
    } while (integer != 0);

    if (negative == TRUE)
    {
       -- current;
       * current = '-';
    }

    -- size;
    if (& buffer [sizeof (buffer)] - current < size)
      size = & buffer [sizeof (buffer)] - current;
   
    memory_copy (string, current, size);

    string [size] = '\0';

    return string;
}

    
char *
string_from_unsigned (char * string, Size size, unsigned integer, Size base)
{
    char buffer [64];
    char * current = & buffer [sizeof (buffer)];

    if (base >= 36 ||
	base <= 1)
      base = 10;

    if (integer <= 0)
    {
       -- current;
       * current = '0';
    }
    else
    do
    {
       int remainder = integer % base;

       integer /= base;
    
       -- current;
       if (remainder < 10)
	 * current = (remainder + '0');
       else
	 * current = (remainder - 10 + 'a');
    } while (integer > 0);
                  
    -- size;
    if (& buffer [sizeof (buffer)] - current < size)
      size = & buffer [sizeof (buffer)] - current;

    memory_copy (string, current, size);

    string [size] = '\0';

    return string;
}

int
string_to_integer (const char * string)
{
   int result = 0;
   Bool negative = FALSE;

   if (* string == '-')
   {
      ++ string;

      negative = TRUE;
   }

   while (* string >= '0' &&
	  * string <= '9')
   {
      result = result * 10 + (* string - '0');
      ++ string;
   }

   if (negative == TRUE)
     result = - result;

   return result;
}

