#include <utility/string.h>
#include <utility/memory.h>
#include <linker/pcom.h>
#include <linker/import.h>
#include <linker/export.h>
#include <linker/section.h>
#include <linker/constructor.h>
#include <linker/component.h>
#include <linker/linker.h>
#include <xcom/xcom.h>

Bool
xcom_load (Component ** result,
           uint8 * memory,
           Size memorySize)
{
    Component * component;
    XComHeader * header = (XComHeader *) memory;
    XComSection * sectionMap = (XComSection *) & header [1],
                * stringsSection = NULL,
                * interfaceSection = NULL,
                * currentSection;
    const char * strings;
    PCom pcom;

    if (memorySize < header -> fileSize ||
        memory_compare (header -> signature, XCOM_SIGNATURE, XCOM_SIGNATURE_LENGTH) != COMPARE_EQUAL ||
        (header -> referenceAttributes & XCOM_REFERENCE_ATTRIBUTE_SIZE) != XCOM_HOST_REFERENCE_ATTRIBUTE_SIZE ||
        (header -> referenceAttributes & XCOM_REFERENCE_ATTRIBUTE_BIG_ENDIAN) != XCOM_HOST_REFERENCE_ATTRIBUTE_BIG_ENDIAN)
      return FALSE;

    for (currentSection = sectionMap;
         currentSection < & sectionMap [header -> sections];
         ++ currentSection)
    {
        switch (currentSection -> type)
        {
        case XCOM_SECTION_TYPE_STRINGS:
             stringsSection = currentSection;

             break;

        case XCOM_SECTION_TYPE_INTERFACE:
             interfaceSection = currentSection;

             break;

        default:
             break;
        }
    }

    if (stringsSection == NULL ||
        interfaceSection == NULL)
      return FALSE;

    strings = (char *) & memory [stringsSection -> offset];

    if (interfaceSection -> dataIndex == 0 ||
        string_equal (& strings [interfaceSection -> dataIndex], "pos") == FALSE ||
        pcom_create (& pcom, & memory [interfaceSection -> offset], (Size) interfaceSection -> size) == FALSE)
      return FALSE;

    if (component_create (& component, pcom.name) == FALSE)
    {
       pcom_destroy (& pcom);
       
       return FALSE;
    }

    for (currentSection = sectionMap;
         currentSection < & sectionMap [header -> sections];
         ++ currentSection)
    {
        XComSection * relocationSection;
        Size relocationCount = 0;
        SectionCapabilities capabilities = 0;
        uint8 * data = NULL;

        if (currentSection -> type != XCOM_SECTION_TYPE_PROGRAM)
          continue;

        if (currentSection -> memoryAttributes & XCOM_MEMORY_ATTRIBUTE_READ)
          capabilities |= SECTION_CAPABILITY_READ;

        if (currentSection -> memoryAttributes & XCOM_MEMORY_ATTRIBUTE_WRITE)
          capabilities |= SECTION_CAPABILITY_WRITE;

        if (currentSection -> memoryAttributes & XCOM_MEMORY_ATTRIBUTE_EXECUTE)
          capabilities |= SECTION_CAPABILITY_EXECUTE;

        if (currentSection -> sectionAttributes & XCOM_SECTION_ATTRIBUTE_CONTENTS)
          data = & memory [currentSection -> offset];

        for (relocationSection = sectionMap;
             relocationSection < & sectionMap [header -> sections];
             ++ relocationSection)
        {
            if (relocationSection -> type == XCOM_SECTION_TYPE_RELOCATIONS &&
                & sectionMap [relocationSection -> dataIndex - 1] == currentSection)
            {
               relocationCount = relocationSection -> size / sizeof (XComRelocation);

               break;
            }
        }

        if (section_create (NULL, 
                            component, 
                            & strings [currentSection -> nameIndex],
                            capabilities,
                            data,
                            (Size) (1 << currentSection -> alignment),
                            (Size) currentSection -> size,
                            relocationCount) == FALSE)
          goto failure;
    }

    if (pcom_add_imports (& pcom, component) == FALSE ||
        pcom_add_exports (& pcom, component) == FALSE ||
        pcom_add_constructors (& pcom, component) == FALSE ||
        pcom_add_destructors (& pcom, component) == FALSE)
      goto failure;

    for (currentSection = sectionMap;
         currentSection < & sectionMap [header -> sections];
         ++ currentSection)
    {
        Section * section = NULL;
        XComSymbol * symbols,
                   * currentSymbol;

        if (currentSection -> type != XCOM_SECTION_TYPE_SYMBOLS ||
            currentSection -> dataIndex == 0)
          continue;

        symbols = (XComSymbol *) & memory [currentSection -> offset];
        section = component_find_section (component,
                                          & strings [sectionMap [currentSection -> dataIndex - 1].nameIndex]);
           
        for (currentSymbol = symbols;
             currentSymbol < & symbols [currentSection -> size / sizeof (XComSymbol)];
             ++ currentSymbol)
        {
            PComSectionType sectionType;
            const char * sectionName,
                       * symbolAlias;
            Interface * interface;
            ExportType exportType;

            if (pcom_find_symbol (& pcom,
                                  & strings [currentSymbol -> nameIndex],
                                  & sectionType,
                                  & sectionName,
                                  & symbolAlias,
                                  (void **) & interface) == FALSE)
              continue;

            if (sectionType != PCOM_SECTION_TYPE_EXPORTS &&
                sectionType != PCOM_SECTION_TYPE_CONSTRUCTORS &&
                sectionType != PCOM_SECTION_TYPE_DESTRUCTORS)
              goto failure;

            switch (currentSymbol -> type)
            {
            case XCOM_SYMBOL_TYPE_GENERIC:
                 exportType = EXPORT_TYPE_UNKNOWN;

                 break;

            case XCOM_SYMBOL_TYPE_DATA:
                 exportType = EXPORT_TYPE_DATA;

                 break;

            case XCOM_SYMBOL_TYPE_FUNCTION:
                 exportType = EXPORT_TYPE_FUNCTION;

                 break;

            default:
                 goto failure;
            }

            if (export_create (NULL,
                               (sectionType == PCOM_SECTION_TYPE_EXPORTS) ? interface : NULL,
                               symbolAlias,
                               section,
                               exportType,
                               (Address) currentSymbol -> offset,
                               (Size) currentSymbol -> size) == FALSE)
              goto failure;
        }
    }

    for (currentSection = sectionMap;
         currentSection < & sectionMap [header -> sections];
         ++ currentSection)
    {
        XComRelocation * relocations,
                       * currentRelocation;
        Section * target;

        if (currentSection -> type != XCOM_SECTION_TYPE_RELOCATIONS)
          continue;
      
        target = component_find_section (component,
                                         & strings [sectionMap [currentSection -> dataIndex - 1].nameIndex]);

        relocations = (XComRelocation *) & memory [currentSection -> offset];

        for (currentRelocation = relocations;
             currentRelocation < & relocations [currentSection -> size / sizeof (XComRelocation)];
             ++ currentRelocation)
        {
            PComSectionType sectionType;
            const char * sectionName,
                       * symbolAlias;
            Import * import; 
            RelocationType relocationType;
            ImportType importType;

            if (pcom_find_symbol (& pcom,
                                  & strings [currentRelocation -> nameIndex],
                                  & sectionType,
                                  & sectionName,
                                  & symbolAlias,
                                  NULL) == FALSE)
            {
               import = component_find_import (component, NULL, & strings [currentRelocation -> nameIndex]);

               if (import == NULL)
                 goto failure;
            }
            else
              import = component_find_import (component, sectionName, symbolAlias);

            switch (currentRelocation -> symbolType)
            {
            case XCOM_SYMBOL_TYPE_GENERIC:
                 importType = IMPORT_TYPE_UNKNOWN;

                 break;

            case XCOM_SYMBOL_TYPE_DATA:
                 importType = IMPORT_TYPE_DATA;

                 break;

            case XCOM_SYMBOL_TYPE_FUNCTION:
                 importType = IMPORT_TYPE_FUNCTION;

                 break;

            case XCOM_SYMBOL_TYPE_SECTION:
                 importType = IMPORT_TYPE_SECTION;

                 break;

            default:
                 goto failure;
            }

            if (import == NULL)
            {
                if (import_create (& import,
                                   component,
                                   sectionName,
                                   symbolAlias,
                                   importType) == FALSE)
                  goto failure;
            }                    
           
            if (importType != import -> type)
              goto failure;

            switch (currentRelocation -> relocationType)
            {
            case XCOM_RELOCATION_TYPE_IA32_ABSOLUTE_16:
                 relocationType = RELOCATION_TYPE_I386_ABSOLUTE_16;

                 break;

            case XCOM_RELOCATION_TYPE_IA32_ABSOLUTE_32:
                 relocationType = RELOCATION_TYPE_I386_ABSOLUTE_32;

                 break;
                 
            case XCOM_RELOCATION_TYPE_IA32_RELATIVE_8:
                 relocationType = RELOCATION_TYPE_I386_RELATIVE_8;

                 break;

            case XCOM_RELOCATION_TYPE_IA32_RELATIVE_16:
                 relocationType = RELOCATION_TYPE_I386_RELATIVE_16;

                 break;

            case XCOM_RELOCATION_TYPE_IA32_RELATIVE_32:
                 relocationType = RELOCATION_TYPE_I386_RELATIVE_32;

                 break;

            default:
                 goto failure;
            }

            relocation_create (& target -> relocations [currentRelocation - relocations],
                               relocationType,
                               import,
                               (Address) currentRelocation -> offset,
                               0);
        }
    }

    if (result != NULL)
      * result = component;

    pcom_destroy (& pcom);

    return TRUE;
     
failure:
  
    pcom_destroy (& pcom);

    component_destroy (component);

    return FALSE;
}

Bool
xcom_boot (Component ** result, uint8 * memory, Size memorySize)
{
    Component * component;
    XComHeader * header = (XComHeader *) memory;
    XComSection * sectionMap = (XComSection *) & header [1],
                * stringsSection = NULL,
                * interfaceSection = NULL,
                * currentSection;
    PComHeader * pcomHeader;
    PComSection * constructorsSection,
                * interfaceStringsSection;
    PComSymbol * symbols,
               * currentSymbol;
    const char * strings,
               * interfaceStrings;

    if (memorySize < header -> fileSize ||
        memory_compare (header -> signature, XCOM_SIGNATURE, XCOM_SIGNATURE_LENGTH) != COMPARE_EQUAL ||
        (header -> referenceAttributes & XCOM_REFERENCE_ATTRIBUTE_SIZE) != XCOM_HOST_REFERENCE_ATTRIBUTE_SIZE ||
        (header -> referenceAttributes & XCOM_REFERENCE_ATTRIBUTE_BIG_ENDIAN) != XCOM_HOST_REFERENCE_ATTRIBUTE_BIG_ENDIAN)
      return FALSE;

    for (currentSection = sectionMap;
         currentSection < & sectionMap [header -> sections];
         ++ currentSection)
    {
        switch (currentSection -> type)
        {
        case XCOM_SECTION_TYPE_STRINGS:
             stringsSection = currentSection;

             break;

        case XCOM_SECTION_TYPE_INTERFACE:
             interfaceSection = currentSection;

             break;

        case XCOM_SECTION_TYPE_PROGRAM:
             if ((currentSection -> sectionAttributes & (XCOM_SECTION_ATTRIBUTE_COMPLETE |
                                                         XCOM_SECTION_ATTRIBUTE_CONTENTS |
                                                         XCOM_SECTION_ATTRIBUTE_VMA)) != (XCOM_SECTION_ATTRIBUTE_COMPLETE |
                                                                                          XCOM_SECTION_ATTRIBUTE_CONTENTS |
                                                                                          XCOM_SECTION_ATTRIBUTE_VMA))
               return FALSE;

             break;

        default:
             break;
        }
    }

    if (stringsSection == NULL ||
        interfaceSection == NULL)
      return FALSE;

    strings    = (char *) & memory [stringsSection -> offset];
    pcomHeader = (PComHeader *) & memory [interfaceSection -> offset];

    if (interfaceSection -> dataIndex == 0 ||
        string_equal (& strings [interfaceSection -> dataIndex], "pos") == FALSE ||
        pcom_validate (pcomHeader) == FALSE)
      return FALSE;

    interfaceStringsSection = (PComSection *) & pcomHeader [1] + pcomHeader -> stringsSection;
    interfaceStrings        = (const char *) ((uint8 *) pcomHeader + interfaceStringsSection -> offset);

    if (pcomHeader -> constructorsSection == PCOM_SECTION_NUMBER_NULL)
      return FALSE;

    constructorsSection = (PComSection *) & pcomHeader [1] + pcomHeader -> constructorsSection;
    symbols             = (PComSymbol *) ((uint8 *) pcomHeader + constructorsSection -> offset);

    for (currentSymbol = symbols;
         currentSymbol < & symbols [constructorsSection -> symbolCount];
         ++ currentSymbol)
    {
        ConstructorFunction * constructorFunction;
  
        if (currentSymbol -> name == PCOM_STRING_NULL)
          return FALSE;

        constructorFunction = (ConstructorFunction *) xcom_find_symbol (header,
                                                                        sectionMap,
                                                                        memory,
                                                                        strings,
                                                                        & interfaceStrings [currentSymbol -> name]);
        if (constructorFunction == NULL ||
            (* constructorFunction) () == FALSE)
          return FALSE;
    }

    linker_activate (NULL);

    if (xcom_load (& component, memory, memorySize) == FALSE)
      return FALSE;

    component -> state = COMPONENT_STATE_READY;

    if (result != NULL)
      * result = component;
    
    return TRUE;
}

void *
xcom_find_symbol (XComHeader * header, XComSection * sectionMap, uint8 * memory, const char * strings, const char * name)
{
    XComSection * currentSection;
    
    for (currentSection = sectionMap;
         currentSection < & sectionMap [header -> sections];
         ++ currentSection)
    {
        XComSymbol * symbols,
                   * currentSymbol;
                   
        if (currentSection -> type != XCOM_SECTION_TYPE_SYMBOLS ||
            currentSection -> dataIndex == 0)
          continue;

        symbols = (XComSymbol *) & memory [currentSection -> offset];
            
        for (currentSymbol = symbols;
             currentSymbol < & symbols [currentSection -> size / sizeof (XComSymbol)];
             ++ currentSymbol)
        {
            if (string_equal (& strings [currentSymbol -> nameIndex], name) == TRUE)
              return & memory [sectionMap [currentSection -> dataIndex - 1].offset] + currentSymbol -> offset;
        }
    }

    return NULL;
}
         
         
