(defstruct (state (:print-function state-print))
     (index     nil     :type (or fixnum null))
     (type      nil     :type symbol)
     (token     nil     :type (or string null))
     (data      nil     :type (or simple-vector null)))

(defun state-next (state code)
     (if (null (state-data state))
         nil
         (elt (state-data state) code)))

(defun set-state-next (state code value)
     (if (null (state-data state))
         (setf (state-data state) (make-array '(256) :initial-element nil)))
     (setf (elt (state-data state) code) value))

(defsetf state-next set-state-next)

(defstruct (rdt (:print-function rdt-print))
     (states         #()     :type list)
     (displacements  #()     :type array)
     (transitions    #()     :type array)
     (check          #()     :type array))

(defun state-print (state stream level)
     (if (state-index state)
         (case (state-type state)
               (unbounded (format stream "SU#<~A {" (state-index state)))
               (bounded   (format stream "SB#<~A {" (state-index state)))
               (t         (format stream "S#<~A {" (state-index state))))
         (format stream "S#<{"))
     (if (state-token state)
         (format stream "\"~A\"" (state-token state))
         (write (system::get-lisp-obj-address state) :stream stream :radix nil :base 16))
     (format stream "}>")
     level)

(defun rdt-print (rdt stream level)
     (format stream "states(~A):        ~A~%" (length (rdt-states rdt)) (rdt-states rdt))
     (format stream "displacements(~A): ~A~%" (length (rdt-displacements rdt)) (rdt-displacements rdt))
     (format stream "transitions(~A):   ~A~%" (length (rdt-transitions rdt)) (rdt-transitions rdt))
     (format stream "check(~A):         ~A~%" (length (rdt-check rdt)) (rdt-check rdt))
     level)

(defun lex-end (continuation)
     (if (or (null continuation)
             (and (consp (first continuation))
                  (eq (first (first continuation)) 'token)))
         t
         nil))

(defun lex-link (lex pattern continuation)
     (typecase pattern
           (fixnum
                      (let ((old-continuation (state-next lex pattern)))
                           (if (null old-continuation)
                               (setf (state-next lex pattern)
                                     (lex-link (make-state :type (if (lex-end continuation) 'bounded nil))
                                               (first continuation)
                                               (rest continuation)))
                               (progn
                                      (if (eq (state-type old-continuation) 'unbounded)
                                          (return-from lex-link lex))
                                      (if (lex-end continuation)
                                          (setf (state-type old-continuation) 'bounded))
                                      (lex-link old-continuation (first continuation) (rest continuation)))))
                      lex)
           (base-char
                      (lex-link lex (char-code pattern) continuation))
           (string
                      (lex-link lex (map 'list #'identity pattern) continuation))
           (cons
                      (case (first pattern)
                            (-   
                                 (let ((new-continuation (if continuation `((merge ,continuation)) nil)))
                                      (dotimes (offset (- (1+ (char-code (third pattern))) (char-code (second pattern))) lex)
				               (setf lex (lex-link lex (+ (char-code (second pattern)) offset) new-continuation)))))
                            (not
                                 (let ((new-continuation (if continuation `((merge ,continuation)) nil)))
                                      (dotimes (code 256 lex)
                                               (unless (find-if #'(lambda (char) (eq code (char-code char))) (rest pattern))
                                                       (setf lex (lex-link lex code new-continuation))))))
                            (?
                                 (lex-link (lex-link lex (rest pattern) continuation)
                                           (first continuation) 
                                           (rest continuation)))
                            (or
                                 (let ((new-continuation (if continuation `((merge ,continuation)) nil)))
                                      (dolist (part (rest pattern) lex)
                                              (setf lex (lex-link lex part new-continuation)))))
                            (+ 
                                 (lex-link (lex-link lex (rest pattern) `((cycle ,@(rest pattern)) ,@continuation))
                                           (rest pattern)
                                           continuation))
                            (*
                                 (lex-link (lex-link lex (rest pattern) `((cycle ,@(rest pattern)) ,@continuation))
                                           (first continuation) 
                                           (rest continuation)))
                            (cycle
                                 (lex-link lex (rest pattern) `((self ,lex) ,@continuation)))
                            (self
                                 (let ((self-lex (lex-link (second pattern) (first continuation) (rest continuation))))
                                      (setf (state-type self-lex) (if (lex-end continuation) 'unbounded nil))
                                      self-lex))
                            (merge
                                 (if (state-p (second pattern))
                                     (second pattern)
                                     (setf (second pattern) (lex-link lex (first (second pattern)) (rest (second pattern))))))
                            (shared
                                 (setf (second pattern) (lex-link lex (first continuation) (rest continuation))))
                            (token
                                 (setf (state-token lex) (second pattern))
                                 lex)
                            (t   
                                 (lex-link lex (first pattern) (append (rest pattern) continuation)))))
           (nil
                      lex)))

(defun lex-states (root)
     (let ((states  '())
           (counter 0))
          (labels ((traverse-state (state)
                        (setf (state-index state) counter)
			(incf counter)
                        (push state states)
                        (if (simple-vector-p (state-data state))
                            (let ((transitions (map 'list #'identity (remove-if-not #'state-p (state-data state)))))
                                 (dolist (transition transitions)
                                         (if (null (state-index transition))
                                             (traverse-state transition)))))))
                  (traverse-state root)
		  (reverse states))))
                  
(defun lex-make (patterns)
     (lex-states (lex-link (make-state) (first patterns) (rest patterns))))

(defun rdt-try-match (dst src offset)
       (let ((matches (min (- (length dst) offset) (length src))))
            (dotimes (i matches)
                     (if (and (elt dst (+ offset i))
                              (elt src i))
                         (return-from rdt-try-match nil)))
            t))

(defun rdt-find-match (dst src)
      (dotimes (i (length dst))
               (if (rdt-try-match dst src i)
                   (return-from rdt-find-match i)))
      nil)

(defun rdt-insert (dst check src offset state)
      (dotimes (i (length src))
               (when (elt src i)
                     (setf (elt check (+ offset i)) state)
                     (setf (elt dst (+ offset i)) (elt src i))))
      dst)

(defun rdt-compress (states)
     (let* ((size          (* 256 (length states)))
            (displacements (make-array `(,(length states)) :initial-element nil))
            (transitions   (make-array `(,size) :initial-element nil))
            (check         (make-array `(,size) :initial-element nil)))
           (dolist (state states)
                   (if (simple-vector-p (state-data state))
                       (let ((offset (rdt-find-match transitions (state-data state))))
                            (rdt-insert transitions check (state-data state) offset state)
                            (setf (elt displacements (state-index state)) offset))))
           (let ((new-size (1+ (position-if-not #'null transitions :from-end t))))
                (make-rdt :states        states
                          :displacements displacements
                          :transitions   (subseq transitions 0 new-size)
                          :check         (subseq check 0 new-size)))))

(defmacro lexer (&rest patterns)
     `(rdt-compress (lex-make '((or ,@patterns)))))

(defun llex-type (size)
     (cond
           ((>= size 65536) "unsigned int")
           ((>= size 256)   "unsigned short")
           (t               "unsigned char")))

(defun llex-enum (names values stream)
      (format stream "enum~%{~%")
      (mapc #'(lambda (name value)
                      (if (not (eq (first (last names)) name))
                          (format stream "    ~A = ~A,~%" name value)
                          (format stream "    ~A = ~A~%" name value)))
            names values)
      (format stream "};~%~%"))

(defun llex-struct (struct names types stream)
      (format stream "typedef struct~%{~%")
      (mapc #'(lambda (name type)
                      (format stream "    ~A ~A;~%" type name))
            names types)
      (format stream "} ~A;~%~%" struct))

(defun llex-typedef (name type stream)
      (format stream "typedef ~A ~A;~%~%" type name))

(defun llex-prototype (return-type name arguments stream)
      (format stream "extern ~A ~A ~A;~%~%" return-type name arguments))

(defun llex-array (values default stream)
      (format stream "{ ")
      (dolist (value (subseq values 0 (1- (length values))))
              (if (null value)
                  (setf value default))
              (format stream "~A, " value))
      (let ((value (first (last values))))
           (if (null value)
               (setf value default))
           (format stream "~A }" value)))

(defconstant llex-disclaimer "/* This file was automatically generated by LLex. Do not edit. */~%~%")

(defun llex-header (rdt stream)
     (format stream "#ifndef __PIDL_LLEXER_H__~%#define __PIDL_LLEXER_H__~%~%#include <stdlib.h>~%#include <stdio.h>~%~%")
     (format stream llex-disclaimer)
     (llex-enum '("LLEX_STATES" "LLEX_TRANSITIONS" "LLEX_BUFFER_SIZE")
                (list (length (rdt-states rdt))
                      (length (rdt-transitions rdt))
                      1024)
                stream)
     (llex-enum '("LLEX_TOKEN_EOF" "LLEX_TOKEN_ERROR") '(0 1) stream)
     (llex-typedef "LLexState" (llex-type (length (rdt-states rdt))) stream)
     (llex-typedef "LLexToken" "unsigned int" stream)
     (llex-typedef "LLexBool" "char" stream)
     (llex-struct "LLexer" '("inputBuffer [LLEX_BUFFER_SIZE]" "inputChar" "inputEnd" "inputFile") '("char" "char *" "char *" "FILE *") stream)
     (llex-prototype "void" "llexer_create" "(LLexer *, FILE *)" stream)
     (llex-prototype "LLexToken" "llexer_lex" "(LLexer *, char *, size_t)" stream)
     (llex-prototype "LLexBool" "llexer_find" "(LLexer *, char)" stream)
     (format stream "#endif /* __PIDL_LLEXER_H__ */~%~%"))

(defun llex-body (rdt header stream)
     (format stream "#include \"~A\"~%~%" header)
     (format stream llex-disclaimer)
     (format stream "static const size_t llexDisplacements [LLEX_STATES] = ")
     (llex-array (map 'list #'identity (rdt-displacements rdt)) 0 stream)
     (format stream ";~%static const LLexState llexCheck [LLEX_TRANSITIONS] = ")
     (llex-array (map 'list
                      #'(lambda (state)
                                (if state (state-index state) "~0"))
                      (rdt-check rdt))
                 "~0" stream)
     (format stream ";~%static const LLexState llexTransitions [LLEX_TRANSITIONS] = ")
     (llex-array (map 'list 
                      #'(lambda (state)
                                (if state 
                                    (state-index state) 
                                    0))
                      (rdt-transitions rdt))
                 0 stream)                     
     (format stream ";~%static const LLexBool llexEnd [LLEX_STATES] = ")
     (llex-array (mapcar #'(lambda (state)
                                   (if (or (state-type state) (state-token state))
                                       1 
                                       0))
                         (rdt-states rdt))
                 0 stream)
     (format stream ";~%static const LLexToken llexTokens [LLEX_STATES] = ")
     (llex-array (mapcar #'state-token (rdt-states rdt)) 1 stream)
     (format stream ";~%~%
void
llexer_create (LLexer * lexer, FILE * inputFile)
{
    lexer -> inputChar = & lexer -> inputBuffer [sizeof (lexer -> inputBuffer)];
    lexer -> inputEnd  = & lexer -> inputBuffer [sizeof (lexer -> inputBuffer)];
    lexer -> inputFile = inputFile;
}

static size_t
llexer_read (LLexer * lexer)
{
    size_t count = fread (lexer -> inputBuffer,
                          1,
                          LLEX_BUFFER_SIZE,
                          lexer -> inputFile);

    lexer -> inputChar = lexer -> inputBuffer;
    lexer -> inputEnd  = & lexer -> inputBuffer [count];

    return count;
}

LLexToken
llexer_lex (LLexer * lexer, char * outputChar, size_t outputSize)
{
    LLexState currentState = 0;
    char * outputEnd       = & outputChar [outputSize - 1],
         * inputChar       = lexer -> inputChar,
         * inputEnd        = lexer -> inputEnd;
    size_t nextIndex;

    if (inputChar >= inputEnd &&
        llexer_read (lexer) <= 0)
      return LLEX_TOKEN_EOF;

    inputChar = lexer -> inputChar;
    inputEnd  = lexer -> inputEnd;

    do
    {
        if (inputChar >= inputEnd &&
            llexer_read (lexer) <= 0)
          return LLEX_TOKEN_ERROR;

        inputChar = lexer -> inputChar;
        inputEnd  = lexer -> inputEnd;

        if (outputEnd - outputChar < inputEnd - inputChar)
          inputEnd = & inputChar [outputEnd - outputChar];

        nextIndex = llexDisplacements [currentState] + (size_t) * inputChar;

        while (llexCheck [nextIndex] == currentState)
        {
            * outputChar = * inputChar;
            ++ inputChar;
            ++ outputChar;
    
            currentState = llexTransitions [nextIndex];
            
            if (inputChar >= inputEnd)
              break;
              
            nextIndex = llexDisplacements [currentState] + (size_t) * inputChar;
        }
    } while (inputChar >= inputEnd);

    lexer -> inputChar = inputChar;

    if (llexEnd [currentState] == 0)
      return LLEX_TOKEN_ERROR;

    * outputChar = '\\0';

    return llexTokens [currentState];
}

LLexBool
llexer_find (LLexer * lexer, char seekChar)
{
    char * inputChar = lexer -> inputChar,
         * inputEnd  = lexer -> inputEnd;

    if (inputChar >= inputEnd &&
        llexer_read (lexer) <= 0)
      return LLEX_TOKEN_EOF;

    inputChar = lexer -> inputChar;
    inputEnd  = lexer -> inputEnd;

    do
    {
        if (inputChar >= inputEnd &&
            llexer_read (lexer) <= 0)
          return 0;

        while (inputChar < inputEnd)
        {
           if (* inputChar == seekChar)
             break;

           ++ inputChar;
        }
    } while (inputChar >= inputEnd);

    lexer -> inputChar = inputChar + 1;
    lexer -> inputEnd  = inputEnd;
    
    return 1;
}

     ~%"))

(defun llex-generate (rdt file)
     (with-open-file (stream (make-pathname :name (concatenate 'string file ".h"))
                             :if-exists         :rename-and-delete
                             :if-does-not-exist :create
                             :direction         :output)
                     (llex-header rdt stream))
     (with-open-file (stream (make-pathname :name (concatenate 'string file ".c"))
                             :if-exists         :rename-and-delete
                             :if-does-not-exist :create
                             :direction         :output)
                     (llex-body rdt (concatenate 'string file ".h") stream))
     rdt)
 
