#include <string.h>
#include "error.h"
#include "string.h"
#include "namespace.h"

int
namespace_create (Namespace * namespace)
{
    namespace -> buckets         = NULL;
    namespace -> bucketCount     = 0;
    namespace -> definitionCount = 0;

    return namespace_grow (namespace, NAMESPACE_INITIAL_BUCKETS);
}

void
namespace_destroy (Namespace * namespace)
{
    namespace_clear (namespace);
}

void
namespace_clear (Namespace * namespace)
{
    List * currentBucket;

    if (namespace -> bucketCount <= 0)
      return;

    for (currentBucket = namespace -> buckets;
         currentBucket < & namespace -> buckets [namespace -> bucketCount];
         ++ currentBucket)
    {
        while (list_empty (currentBucket) == 0)
           definition_destroy ((Definition *) list_front (currentBucket));
    }

    free (namespace -> buckets);

    namespace -> buckets         = NULL;
    namespace -> bucketCount     = 0;
    namespace -> definitionCount = 0;
}

Definition *
namespace_find (const Namespace * namespace, const char * name)
{
    size_t hash          = string_hash (name, namespace -> bucketCount);
    ListIterator position;

    for (position = list_begin (& namespace -> buckets [hash]);
         position != list_end (& namespace -> buckets [hash]);
         position = list_next (position))
    {
        Definition * definition = (Definition *) position;

        if (strcmp (name, definition -> name) == 0)
          return definition;
    }

    return NULL;
}

Definition *
namespace_insert (Namespace * namespace, const char * name)
{
    size_t hash;
    Definition * definition;

    definition = namespace_find (namespace, name);
    if (definition != NULL)
      return definition;

    if (namespace -> bucketCount - namespace -> definitionCount <= NAMESPACE_EMPTY_THRESHOLD)
    {
       if (namespace_grow (namespace, namespace -> definitionCount * NAMESPACE_GROWTH_SCALE +
                                      NAMESPACE_GROWTH_OFFSET) != 0)
         pidl_error ("Failed to grow namespace at size %u", namespace -> definitionCount);
    }

    hash       = string_hash (name, namespace -> bucketCount);
    definition = definition_create (& namespace -> buckets [hash],
                                    0,
                                    name);

    if (definition == NULL)
      pidl_error ("Failed to insert definition %s into namespace", name);

    return definition;
}

void
namespace_remove (Namespace * namespace, Definition * definition)
{
    definition_destroy (definition);

    -- namespace -> definitionCount;
}

int
namespace_grow (Namespace * namespace, size_t newSize)
{
    List * newBuckets,
         * currentBucket;
    
    if (newSize < NAMESPACE_INITIAL_BUCKETS)
      newSize = NAMESPACE_INITIAL_BUCKETS;

    newBuckets = (List *) malloc (newSize * sizeof (List));
    if (newBuckets == NULL)
      return -1;

    for (currentBucket = newBuckets;
         currentBucket < & newBuckets [newSize];
         ++ currentBucket)
      list_clear (currentBucket);

    if (namespace -> definitionCount < 0)
    {
        if (namespace -> bucketCount > 0 &&
            namespace -> buckets != NULL)
          free (namespace -> buckets);

        namespace -> buckets     = newBuckets;
        namespace -> bucketCount = newSize;

        return 0;
    }

    for (currentBucket = namespace -> buckets;
         currentBucket < & namespace -> buckets [namespace -> bucketCount];
         ++ currentBucket)
    {
        ListIterator position = list_begin (currentBucket);
        
        while (position != list_end (currentBucket))
        {
            Definition * definition = (Definition *) position;
            size_t       hash       = string_hash (definition -> name, newSize);

            position = list_next (position);

            list_move_to_back (& newBuckets [hash], & definition -> definitionList);
        }
    }
    
    free (namespace -> buckets);

    namespace -> buckets     = newBuckets;
    namespace -> bucketCount = newSize;

    return 0;
}

size_t
namespace_assign_offsets (Namespace * namespace)
{
    size_t offset = 1;
    List * currentBucket;

    for (currentBucket = namespace -> buckets;
         currentBucket < & namespace -> buckets [namespace -> bucketCount];
         ++ currentBucket)
    {
        ListIterator position;
        
        for (position = list_begin (currentBucket);
             position != list_end (currentBucket);
             position = list_next (position))
        {
           Definition * definition = (Definition *) position;

           definition -> offset = offset;

           offset += strlen (definition -> name) + 1;
        }
    }

    return offset;
}

int
namespace_write (Namespace * namespace, FILE * file)
{
    List * currentBucket;

    fputc ('\0', file);

    for (currentBucket = namespace -> buckets;
         currentBucket < & namespace -> buckets [namespace -> bucketCount];
         ++ currentBucket)
    {
        ListIterator position;

        for (position = list_begin (currentBucket);
             position != list_end (currentBucket);
             position = list_next (position))
        {
           Definition * definition = (Definition *) position;

           if (fwrite (definition -> name, strlen (definition -> name) + 1, 1, file) != 1)
             return -1;
        }
    }

    return 0;
}

