#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include "string.h"
#include "value.h"

Value *
value_create (ValueType type)
{
       Value * value = (Value *) malloc (sizeof (Value));

       if (value == NULL)
         return NULL;

       if (type == VALUE_TYPE_LIST)
         list_clear (& value -> datum.list);
       else
         memset (value, 0, sizeof (Value));

       value -> type = type;

       return value;
}

void
value_destroy (Value * value)
{
       switch (value -> type)
       {
       case VALUE_TYPE_STRING:
            free (value -> datum.string);

            break;

       case VALUE_TYPE_LIST:
            while (list_empty (& value -> datum.list) == 0)
               value_destroy ((Value *) list_remove_front (& value -> datum.list));

            break;

       default:
            break;
       }
                
       free (value);
}

Value *
value_copy (Value * value)
{
       Value * newValue = value_create (value -> type);
       ListIterator position;

       switch (value -> type)
       {
       case VALUE_TYPE_INTEGER:
            newValue -> datum.integer = value -> datum.integer;

            break;

       case VALUE_TYPE_STRING:
            newValue -> datum.string = string_duplicate (value -> datum.string);

            break;
       
       case VALUE_TYPE_LIST:
            for (position = list_begin (& value -> datum.list);
                 position != list_end (& value -> datum.list);
                 position = list_next (position))
              list_insert_back (& newValue -> datum.list, value_copy ((Value *) position));

            break;

       default:
            break;
       }

       return newValue;
}

Value *
value_read (LLexer * lexer)
{
       static char buffer [4096];

       LLexToken token;

       for (token = llexer_lex (lexer, buffer, sizeof (buffer));
            token != LLEX_TOKEN_EOF && token != LLEX_TOKEN_ERROR;
            token = llexer_lex (lexer, buffer, sizeof (buffer)))
       {
           Value * value;

           switch (token)
           {
           case '(':
                return value_read_list (lexer);
           
           case ')':
                return NULL;

           case 'i':
                value = value_create (VALUE_TYPE_INTEGER);

                value -> datum.integer = atoi (buffer);

                return value;

           case 's':
                value = value_create (VALUE_TYPE_STRING);

                value -> datum.string = string_duplicate (buffer);

                return value;

           case '"':
                value = value_create (VALUE_TYPE_STRING);

                strtok (buffer, "\"");

                value -> datum.string = string_duplicate (buffer + 1);

                return value;

           case ';':
                if (llexer_find (lexer, '\n') == 0)
                  return NULL;

                break;

           default:
                break;
           }
       }

       return NULL;
}

Value *
value_read_list (LLexer * lexer)
{
       Value * list = value_create (VALUE_TYPE_LIST),
             * value;

       for (value = value_read (lexer);
            value != NULL;
            value = value_read (lexer))
         list_insert_back (& list -> datum.list, value);

       return list;
}

void
value_write (Value * value, FILE * file)
{
       ListIterator position;

       switch (value -> type)
       {
       case VALUE_TYPE_LIST:
            fputc ('(', file);

            position = list_begin (& value -> datum.list);

            if (position != list_end (& value -> datum.list))
            {
               value_write ((Value *) position, file);

               position = list_next (position);
            }
            
            while (position != list_end (& value -> datum.list))
            {
               fputc (' ', file);
                
               value_write ((Value *) position, file);
               
               position = list_next (position);
            }

            fputc (')', file);

            break;

       case VALUE_TYPE_INTEGER:
            fprintf (file, "%d", value -> datum.integer);

            break;

       case VALUE_TYPE_STRING:
            fputs (value -> datum.string, file);

            break;

       default:
            break;
       }
}

int
value_match_list (Value * list, const char * pattern, va_list argumentList)
{
       ListIterator position;

       if (* pattern != '(' ||
           list -> type != VALUE_TYPE_LIST)
         return 0;

       ++ pattern;
       
       for (position = list_begin (& list -> datum.list);
            position != list_end (& list -> datum.list);
            position = list_next (position))
       {
           Value * value = (Value *) position;

           while (* pattern == ' ' ||
                  * pattern == 'r')
           {
               if (* pattern == 'r')
                 * va_arg (argumentList, ListIterator *) = position;
              
               ++ pattern;
           }

           switch (* pattern)
           {
           case 'i':
                if (value -> type != VALUE_TYPE_INTEGER)
                  return -1;

                * va_arg (argumentList, int *) = value -> datum.integer;

                break;

           case 's':
                if (value -> type != VALUE_TYPE_STRING)
                  return -1;

                * va_arg (argumentList, const char **) = value -> datum.string;

                break;

           case 'l':
                if (value -> type != VALUE_TYPE_LIST)
                  return -1;

                * va_arg (argumentList, List **) = & value -> datum.list;

           case 'v':
                if (value -> type != VALUE_TYPE_LIST)
                  return -1;

                * va_arg (argumentList, Value **) = value;
                
                break;

           case '\'':
                {
                    const char * endQuote;
                    
                    if (value -> type != VALUE_TYPE_STRING)
                      return -1;
                    
                    endQuote = strchr (pattern + 1, '\'');

                    if (endQuote == NULL)
                      return -1;

                    if (strncmp (value -> datum.string, & pattern [1], endQuote - & pattern [1]) != 0)
                      return -1;

                    pattern = endQuote;

                    break;
                }
                    
           case '_':
                break;

           case '(':
                if (value -> type != VALUE_TYPE_LIST)
                  return -1;

                if (value_match_list (value, pattern, argumentList) != 0)
                  return -1;

                while (* pattern != ')' && 
                       * pattern != '\0')
                  ++ pattern;

                if (* pattern == '\0')
                  return -1;

                break;

           case ')':
                return -1;

           case '*':
                return 0;

           default:
                return -1;
           }

           ++ pattern;
       }

       while (* pattern == ' ' ||
              * pattern == 'r')
       {
             if (* pattern == 'r')
               * va_arg (argumentList, ListIterator *) = position;

             ++ pattern;
       }

       if (* pattern != '*' &&
           * pattern != ')')
         return -1;

       return 0;
}
        
int
value_match (Value * value, const char * pattern, ...)
{
       va_list argumentList;

       va_start (argumentList, pattern);

       if (value_match_list (value, pattern, argumentList) != 0)
       {
          va_end (argumentList);

          return -1;
       }

       va_end (argumentList);

       return 0;
}
       
