#ifndef __XL_ELF_H__
#define __XL_ELF_H__

#include "types.h"

typedef uint8           elf_uint8;
typedef uint16          elf_uint16;
typedef uint32          elf_uint32;

typedef int32           elf_int32;

typedef elf_uint32      elf_file_offset;
typedef elf_uint32      elf_address;

static const char ELF_SIGNATURE [4] = { (char) 0x7F, 'E', 'L', 'F' };

typedef enum
{
        ELF_SIGNATURE_LENGTH = 4
} ELFSignature;

typedef enum
{
        ELF_RESERVED_LENGTH = 16 - ELF_SIGNATURE_LENGTH - 2
} ELFReserved;

typedef enum
{
        ELF_FILE_CLASS_NONE = 0,
        ELF_FILE_CLASS_32   = 1,
        ELF_FILE_CLASS_64   = 2
} ELFFileClass;

static const ELFFileClass ELF_HOST_FILE_CLASS = ELF_FILE_CLASS_32;

typedef enum
{
        ELF_DATA_ENCODING_NONE = 0,
        ELF_DATA_ENCODING_LSB  = 1,
        ELF_DATA_ENCODING_MSB  = 2
} ELFDataEncoding;

static const ELFDataEncoding ELF_HOST_DATA_ENCODING = ELF_DATA_ENCODING_LSB;

typedef enum
{
        ELF_FILE_TYPE_NONE        = 0,
        ELF_FILE_TYPE_RELOCATABLE = 1,
        ELF_FILE_TYPE_EXECUTABLE  = 2,
        ELF_FILE_TYPE_DYNAMIC     = 3,
        ELF_FILE_TYPE_CORE        = 4
} ELFFileType;

typedef enum
{
        ELF_SECTION_TYPE_NONE               = 0,
        ELF_SECTION_TYPE_PROGRAM            = 1,
        ELF_SECTION_TYPE_SYMBOLS            = 2,
        ELF_SECTION_TYPE_STRINGS            = 3,
        ELF_SECTION_TYPE_ADDEND_RELOCATIONS = 4,
        ELF_SECTION_TYPE_HASH_TABLE         = 5,
        ELF_SECTION_TYPE_DYNAMIC            = 6,
        ELF_SECTION_TYPE_NOTE               = 7,
        ELF_SECTION_TYPE_ALLOCATE           = 8,
        ELF_SECTION_TYPE_RELOCATIONS        = 9,
        ELF_SECTION_TYPE_RESERVED           = 10
} ELFSectionType;

typedef enum
{
        ELF_SECTION_ATTRIBUTE_WRITE    = (1 << 0),
        ELF_SECTION_ATTRIBUTE_LOAD     = (1 << 1),
        ELF_SECTION_ATTRIBUTE_EXECUTE  = (1 << 2)
} ELFSectionAttribute;

typedef enum
{
        ELF_SECTION_INDEX_UNDEFINED = 0,
        ELF_SECTION_INDEX_ABSOLUTE  = 0xFFF1,
        ELF_SECTION_INDEX_COMMON    = 0xFFF2
} ELFSectionIndex;

typedef enum
{
        ELF_SYMBOL_BINDING_LOCAL  = (0 << 4),
        ELF_SYMBOL_BINDING_GLOBAL = (1 << 4),
        ELF_SYMBOL_BINDING_WEAK   = (2 << 4)
} ELFSymbolBinding;

typedef enum
{
        ELF_SYMBOL_TYPE_NONE     = 0,
        ELF_SYMBOL_TYPE_DATA     = 1,
        ELF_SYMBOL_TYPE_FUNCTION = 2,
        ELF_SYMBOL_TYPE_SECTION  = 3,
        ELF_SYMBOL_TYPE_FILE     = 4
} ELFSymbolType;

typedef enum
{
        ELF_SYMBOL_ATTRIBUTE_TYPE    = 0xF,
        ELF_SYMBOL_ATTRIBUTE_BINDING = 0xF << 4
} ELFSymbolAttribute;
     
typedef enum
{
        ELF_RELOCATION_TYPE_NONE                 = 0,
        ELF_RELOCATION_TYPE_I386_ABSOLUTE_32     = 1,
        ELF_RELOCATION_TYPE_I386_PC_RELATIVE_32  = 2,
        ELF_RELOCATION_TYPE_I386_GOT_32          = 3,
        ELF_RELOCATION_TYPE_I386_PLT_32          = 4,
        ELF_RELOCATION_TYPE_I386_COPY            = 5,
        ELF_RELOCATION_TYPE_I386_GOT_ENTRY       = 6,
        ELF_RELOCATION_TYPE_I386_PLT_ENTRY       = 7,
        ELF_RELOCATION_TYPE_I386_DYNAMIC         = 8,
        ELF_RELOCATION_TYPE_I386_GOT_RELATIVE    = 9,
        ELF_RELOCATION_TYPE_I386_GOT_PC_RELATIVE = 10
} ELFRelocationType;
        
typedef enum
{
        ELF_MACHINE_TYPE_NONE  = 0,
        ELF_MACHINE_TYPE_M32   = 1,
        ELF_MACHINE_TYPE_SPARC = 2,
        ELF_MACHINE_TYPE_I386  = 3,
        ELF_MACHINE_TYPE_M68K  = 4,
        ELF_MACHINE_TYPE_M88K  = 5,
        ELF_MACHINE_TYPE_I860  = 6,
        ELF_MACHINE_TYPE_MIPS  = 7
} ELFMachineType;
       
typedef enum
{
        ELF_VERSION_NONE    = 0,
        ELF_VERSION_CURRENT = 1
} ELFVersion;

typedef struct
{
        elf_uint8       signature [ELF_SIGNATURE_LENGTH];
        elf_uint8       fileClass;
        elf_uint8       dataEncoding;
        elf_uint8       reserved [ELF_RESERVED_LENGTH];
        elf_uint16      fileType;
        elf_uint16      machineType;
        elf_uint32      version;
        elf_address     entryAddress;
        elf_file_offset programHeaderOffset;
        elf_file_offset sectionHeaderOffset;
        elf_uint32      machineFlags;
        elf_uint16      headerSize;
        elf_uint16      programHeaderSize;
        elf_uint16      programHeaderCount;
        elf_uint16      sectionHeaderSize;
        elf_uint16      sectionHeaderCount;
        elf_uint16      stringSection;
} ELFHeader;

typedef struct
{
        elf_uint32      name;
        elf_uint32      type;
        elf_uint32      attributes;
        elf_address     address;
        elf_file_offset offset;
        elf_uint32      size;
        elf_uint32      link;
        elf_uint32      information;
        elf_uint32      alignment;
        elf_uint32      entrySize;
} ELFSectionHeader;

typedef struct
{
        elf_uint32      name;
        elf_address     value;
        elf_uint32      size;
        elf_uint8       attributes;
        elf_uint8       reserved;
        elf_uint16      section;
} ELFSymbol;

typedef struct
{
        elf_address     offset;
        unsigned        type : 8;
        unsigned        symbol : 16;
        unsigned        reserved : 8;
} ELFRelocation;

typedef struct
{
        elf_address     offset;
        unsigned        type : 8;
        unsigned        symbol : 16;
        unsigned        reserved : 8;
        elf_int32       addend;
} ELFAddendRelocation;

#endif /* __XL_ELF_H__ */

