#include "error.h"
#include "symbol.h"
#include "object.h"
#include "relocation.h"

Relocation *
relocation_create (List * relocations,
                   RelocationType type,
                   size_t offset,
                   Definition * definition)
{
     Relocation * relocation = (Relocation *) malloc (sizeof (Relocation));

     if (relocation == NULL)
       return NULL;

     relocation -> type       = type;
     relocation -> offset     = offset;
     relocation -> definition = definition;
     relocation -> deferred   = 0;
     
     ++ definition -> relocationCount;

     if (relocations != NULL)
       list_insert_back (relocations, & relocation -> relocationList);

     return relocation;
}

void
relocation_destroy (Relocation * relocation)
{
     list_remove (& relocation -> relocationList);

     free (relocation);
}

void
relocation_merge (Relocation * relocation, Section * section)
{
     relocation -> offset += section -> dataSize;

     list_move_to_back (& section -> relocations, & relocation -> relocationList);

     relocation_move (relocation, section, section -> object, relocation -> definition -> name);
}

void
relocation_move (Relocation * relocation, Section * section, Object * object, const char * name)
{
     -- relocation -> definition -> relocationCount;

     relocation -> definition = namespace_find (& object -> namespace, name);

     if (relocation -> definition == NULL)
        relocation -> definition = namespace_insert (& object -> namespace, name, NULL);

     ++ relocation -> definition -> relocationCount;
}

void
relocation_fixup (Relocation * relocation, Section * section, size_t offset)
{
     uint8 * target = & section -> data [relocation -> offset];
     
     switch (relocation -> type)
     {
     case RELOCATION_TYPE_RELATIVE_8:
        * (uint8 *) target += (uint8) offset;
        
        break;

     case RELOCATION_TYPE_ABSOLUTE_16:
     case RELOCATION_TYPE_RELATIVE_16:
        * (uint16 *) target += (uint16) offset;
        
        break;

     case RELOCATION_TYPE_ABSOLUTE_32:
     case RELOCATION_TYPE_RELATIVE_32:
        * (uint32 *) target += (uint32) offset;
        
        break;

     default:
        break;
     }
}        
               
int
relocation_relocate (Relocation * relocation, Section * section, int final)
{
     Symbol * symbol = relocation -> definition -> symbol;
     uint8 * target;
     Address value;

     if (symbol -> attributes & SYMBOL_ATTRIBUTE_IMPORTED)
       return 0;

     if (symbol_resolved (symbol) == 0)
     {
        if (symbol -> type == SYMBOL_TYPE_SECTION)
          return 0;
     
        target = & section -> data [relocation -> offset];
        value  = symbol -> value;

        switch (relocation -> type)
        {
        case RELOCATION_TYPE_RELATIVE_8:
             * (uint8 *) target += (uint8) value;

             break;

        case RELOCATION_TYPE_ABSOLUTE_16:
        case RELOCATION_TYPE_RELATIVE_16:
             * (uint16 *) target += (uint16) value;
             
             break;

        case RELOCATION_TYPE_ABSOLUTE_32:
        case RELOCATION_TYPE_RELATIVE_32:
             * (uint32 *) target += (uint32) value;
             
             break;

        default:
             xl_error ("Unknown relocation type");
        }
      
        relocation_move (relocation, section, section -> object, symbol -> section -> name);

        return 0;     
     }
                
     if (!final)
     {
        switch (relocation -> type)
        {
        case RELOCATION_TYPE_ABSOLUTE_16:
        case RELOCATION_TYPE_ABSOLUTE_32:
            if (symbol_anchoring (symbol))
              return -1;
            
            break;
        
        case RELOCATION_TYPE_RELATIVE_8:
        case RELOCATION_TYPE_RELATIVE_16:
        case RELOCATION_TYPE_RELATIVE_32:
            if (section -> attributes & SECTION_ATTRIBUTE_ANCHORING)
              return -1;

            if ((section -> attributes & SECTION_ATTRIBUTE_ANCHORED) != 0)
              return 0;
            
            if (symbol_anchoring (symbol))
              return -1;
                
            break;

        default:
            xl_error ("Unknown relocation type");
        }
     }

     target = & section -> data [relocation -> offset];
     value  = symbol_value (symbol);

     switch (relocation -> type)
     {
     case RELOCATION_TYPE_ABSOLUTE_16:
        * (uint16 *) target += (uint16) value;

        return 1;

     case RELOCATION_TYPE_ABSOLUTE_32:
        * (uint32 *) target += (uint32) value;
        
        return 1;

     default:
        break;
     }

     if ((section -> attributes & SECTION_ATTRIBUTE_ANCHORED) == 0)
       return 0;
    
     value -= section -> address + relocation -> offset;

     switch (relocation -> type)
     {
     case RELOCATION_TYPE_RELATIVE_8:
        * (uint8 *) target += (uint8) value;

        return 1;

     case RELOCATION_TYPE_RELATIVE_16:
        * (uint16 *) target += (uint16) value;
        
        return 1;
        
     case RELOCATION_TYPE_RELATIVE_32:
        * (uint32 *) target += (uint32) value;
        
        return 1;

     default:
        xl_error ("Unknown relocation type");
     }

     return 0;
}
        
        
