#include <string.h>
#include "error.h"
#include "object.h"
#include "section.h"
#include "relocation.h"

Section *
section_create (Object * object,
                SectionType type,
                SectionAttributes attributes,
                SectionCapabilities capabilities,
                const char * name,
                const char * architecture,
                Address address,
                size_t dataAlignment,
                size_t dataSize,
                uint8 * data)
{
     Section * section = (Section *) malloc (sizeof (Section));

     if (section == NULL)
       return NULL;

     section -> object        = object;
     section -> type          = type;
     section -> attributes    = attributes;
     section -> capabilities  = capabilities;
     section -> name          = name;
     section -> architecture  = architecture;
     section -> address       = address;
     section -> dataAlignment = dataAlignment;
     section -> dataSize      = dataSize;
     section -> data          = data;

     list_clear (& section -> relocations);
     list_clear (& section -> exportedSymbols);

     if (object != NULL)
       list_insert_back (& object -> sections, & section -> sectionList);

     return section;
}

void
section_destroy (Section * section)
{
     list_remove (& section -> sectionList);

     while (list_empty (& section -> relocations) == 0)
        relocation_destroy ((Relocation *) list_front (& section -> relocations));

     while (list_empty (& section -> exportedSymbols) == 0)
        symbol_destroy ((Symbol *) list_front (& section -> exportedSymbols));

     if (section -> data != NULL &&
         section -> dataSize > 0)
       free (section -> data);
     
     if ((section -> attributes & SECTION_ATTRIBUTE_STRIPPED) == 0)
     {
        if (section -> name != NULL)
          definition_dereference (namespace_find (& section -> object -> namespace, section -> name));

        if (section -> architecture != NULL)
          definition_dereference (namespace_find (& section -> object -> namespace, section -> architecture));
     }

     free (section);
}

void
section_strip (Section * section)
{
     if (section -> name != NULL)
       definition_dereference (namespace_find (& section -> object -> namespace, section -> name));

     if (section -> architecture != NULL)
       definition_dereference (namespace_find (& section -> object -> namespace, section -> architecture));

     section -> attributes |= SECTION_ATTRIBUTE_STRIPPED;
}
     
int
section_adjacent (const Section * dst, const Section * src)
{
    if (dst == src ||
        dst -> name == NULL ||
        src -> name == NULL ||
        strcmp (dst -> name, src -> name) != 0)
      return 0;

    if (dst -> type != SECTION_TYPE_PROGRAM ||
        src -> type != SECTION_TYPE_PROGRAM ||
        (dst -> attributes & SECTION_ATTRIBUTE_MASK) != (src -> attributes & SECTION_ATTRIBUTE_MASK) ||
        (dst -> capabilities & SECTION_CAPABILITY_MASK) != (src -> capabilities & SECTION_CAPABILITY_MASK) ||
        ((dst -> attributes & SECTION_ATTRIBUTE_ANCHORED) != 0 &&
         src -> address != dst -> address + dst -> dataSize) ||
        strcmp (dst -> architecture, src -> architecture) != 0)
      xl_error ("Incompatable sections with the same name: %s", dst -> name);
    
    return 1;
}

int
section_merge (Section * dst, Section * src)
{
     ListIterator position;
   
     if (src -> dataAlignment > dst -> dataAlignment)
       dst -> dataAlignment = src -> dataAlignment;
        
     dst -> attributes   |= src -> attributes;
     dst -> capabilities |= src -> capabilities;

     if (dst -> dataSize + src -> dataSize > 0 &&
         (dst -> attributes & SECTION_ATTRIBUTE_CONTENTS) != 0)
     {
        uint8 * newData = (uint8 *) realloc (dst -> data, 
                                             dst -> dataSize + src -> dataSize);
        if (newData == NULL)
          return -1;

        memcpy (& newData [dst -> dataSize], src -> data, src -> dataSize);
        
        dst -> data = newData;
     }

     object_define_symbols (dst -> object, dst, & src -> exportedSymbols);

     position = list_begin (& src -> relocations);
     while (position != list_end (& src -> relocations))
     {
         Relocation * relocation = (Relocation *) position;
      
         position = list_next (position);
         
         relocation_merge (relocation, dst);
     }

     dst -> dataSize += src -> dataSize;

     section_destroy (src);

     return 0;
}    

int
section_move (Section * section, Object * object)
{
     List exportedSymbols;
     ListIterator position;

     list_splice (& exportedSymbols, 
                  list_begin (& section -> exportedSymbols), 
                  list_end (& section -> exportedSymbols));
    
     object_define_symbols (object, section, & exportedSymbols);
     
     for (position = list_begin (& section -> relocations);
          position != list_end (& section -> relocations);
          position = list_next (position))
     {
         Relocation * relocation = (Relocation *) position;
    
         relocation_move (relocation, section, object, relocation -> definition -> name);
     }

     while (list_empty (& exportedSymbols) == 0)
        symbol_destroy ((Symbol *) list_front (& exportedSymbols));

     list_move_to_back (& object -> sections, & section -> sectionList);

     if (section -> name != NULL)
       section -> name = object_translate_string (object, section -> name);
     
     if (section -> architecture != NULL)
       section -> architecture = object_translate_string (object, section -> architecture);

     section -> object = object;

     return 0;
}

void
section_fixup (Section * section, Section * target, size_t offset)
{
     ListIterator position;

     for (position = list_begin (& section -> relocations);
          position != list_end (& section -> relocations);
          position = list_next (position))
     {
        Relocation * relocation = (Relocation *) position;
        
        if (relocation -> definition -> symbol -> type == SYMBOL_TYPE_SECTION &&
            relocation -> definition -> symbol -> section == target)
          relocation_fixup (relocation, section, offset);
     }

     if (section == target)
     for (position = list_begin (& section -> exportedSymbols);
          position != list_end (& section -> exportedSymbols);
          position = list_next (position))
     {
        Symbol * symbol = (Symbol *) position;

        symbol -> value += offset;
     }
}

void
section_relocate (Section * section, int final)
{
     ListIterator position = list_begin (& section -> relocations);

     while (position != list_end (& section -> relocations))
     {
        Relocation * relocation = (Relocation *) position;

        position = list_next (position);

        switch (relocation_relocate (relocation, section, final))
        {
        case 1:
           -- relocation -> definition -> relocationCount;

           relocation_destroy (relocation);
              
           break;

        case -1:
           -- relocation -> definition -> relocationCount;

           relocation -> deferred = 1;

           break;    
        }
     }
}

int
section_read (Section * section, FILE * file)
{
     if (section -> dataSize <= 0 ||
         (section -> type == SECTION_TYPE_PROGRAM &&
          (section -> attributes & SECTION_ATTRIBUTE_CONTENTS) == 0))
       return 0;

     if (section -> data == NULL)
     {
         uint8 * data = (uint8 *) realloc (section -> data,
                                           section -> dataSize);

         if (data == NULL)
           return -1;

         section -> data = data;
     }
     
     if (fread (section -> data, section -> dataSize, 1, file) != 1)
       return -1;

     return 0;
}

int
section_write (const Section * section, FILE * file, int zeroFill)
{
     if (section -> data != NULL &&
         (section -> type != SECTION_TYPE_PROGRAM ||
          (section -> attributes & SECTION_ATTRIBUTE_CONTENTS) != 0) &&
         fwrite (section -> data, section -> dataSize, 1, file) != 1)
       return -1;
     else
     if (zeroFill != 0 &&
         section -> type == SECTION_TYPE_PROGRAM &&
         (section -> attributes & SECTION_ATTRIBUTE_CONTENTS) == 0 &&
         (fseek (file, section -> dataSize - 1, SEEK_CUR) != 0 ||
          fputc (0, file) == EOF))
       return -1;

     return 0;
}

