#include <stdlib.h>
#include <string.h>
#include "xcom.h"
#include "error.h"
#include "object.h"
#include "symbol.h"
#include "section.h"
#include "relocation.h"

static size_t
count_symbols (const List * list)
{
    ListIterator position;
    size_t total = 0;

    for (position = list_begin (list);
         position != list_end (list);
         position = list_next (position))
    {
        Symbol * symbol = (Symbol *) position;

        if (symbol -> attributes & SYMBOL_ATTRIBUTE_STRIPPED)
          continue;
        
        ++ total;
    }

    return total;
}

static size_t
count_relocations (const List * list)
{
    ListIterator position;
    size_t total = 0;

    for (position = list_begin (list);
         position != list_end (list);
         position = list_next (position))
    {
        Relocation * reloc = (Relocation *) position;

        if (reloc -> deferred)
          continue;

        ++ total;
    }

    return total;
}

int
xcom_anchor (Object * object, Address address, int offset, const char * name, int zeroFill)
{
     ListIterator position;
     xc_reference currentOffset = sizeof (XComHeader) + sizeof (XComSection);

     for (position = list_begin (& object -> sections);
          position != list_end (& object -> sections);
          position = list_next (position))
     {
         Section * section = (Section *) position;

         if (section -> attributes & SECTION_ATTRIBUTE_STRIPPED)
           continue;
       
         currentOffset += sizeof (XComSection);

         if (count_symbols (& section -> exportedSymbols))
           currentOffset += sizeof (XComSection);

         if (count_relocations (& section -> relocations))
           currentOffset += sizeof (XComSection);
     } 

     if (count_symbols (& object -> importedSymbols) || count_symbols (& object -> constantSymbols))
       currentOffset += sizeof (XComSection);

     currentOffset += namespace_assign_offsets (& object -> namespace, 0);
    
     for (position = list_begin (& object -> sections);
          position != list_end (& object -> sections);
          position = list_next (position))
     {
        Section * section = (Section *) position;
        xc_reference alignment,
                     alignOffset;

        if (section -> attributes & SECTION_ATTRIBUTE_STRIPPED)
          continue;

        if (section -> type == SECTION_TYPE_PROGRAM)
        {
           alignment   = (xc_reference) (1 << section -> dataAlignment);

           if (zeroFill == 0 &&
               (section -> attributes & SECTION_ATTRIBUTE_CONTENTS) == 0)
           {
              if ((section -> attributes & SECTION_ATTRIBUTE_ANCHORED) == 0 && (name == NULL || strcmp (section -> name, name) == 0))
              {
                 if (offset) 
                   return -1;
           
                 alignOffset = address % alignment;
                 if (alignOffset > 0)
                   address += alignment - alignOffset;
 
                 section -> attributes |= SECTION_ATTRIBUTE_ANCHORED;
                 section -> address = address;
                
                 address += section -> dataSize;
              } 
              
              continue;
           }
 
           alignOffset = currentOffset % alignment;
           if (alignOffset > 0)
             currentOffset += alignment - alignOffset;

           if ((section -> attributes & SECTION_ATTRIBUTE_ANCHORED) == 0 && (name == NULL || strcmp (section -> name, name) == 0))
           {
              section -> attributes |= SECTION_ATTRIBUTE_ANCHORED;

              if (offset)
                 section -> address = address + currentOffset;
              else
              {
                 section -> address = address;
                 
                 address += section -> dataSize;
              }
           }
                
        }

        currentOffset += section -> dataSize;
     }
    
     return 0;
}

int
xcom_write (Object * object, FILE * file, int zeroFill)
{
     XComHeader header;
     XComSection * sectionMap,
                 * currentSection,
                 * targetSection;
     xc_reference currentOffset;
     size_t numSyms = 0;
     ListIterator position;

     memcpy (header.signature, XCOM_SIGNATURE, XCOM_SIGNATURE_LENGTH);

     header.referenceAttributes = XCOM_HOST_REFERENCE_ATTRIBUTE_SIZE |
                                  XCOM_HOST_REFERENCE_ATTRIBUTE_BIG_ENDIAN;
     header.version             = XCOM_VERSION;

     sectionMap = (XComSection *) calloc (3 * list_size (& object -> sections) + 2, sizeof (XComSection));
     if (sectionMap == NULL)
       return -1;

     sectionMap -> type              = XCOM_SECTION_TYPE_STRINGS;
     sectionMap -> sectionAttributes = XCOM_SECTION_ATTRIBUTE_MANDATORY;
     sectionMap -> size              = namespace_assign_offsets (& object -> namespace, 1);

     for (position = list_begin (& object -> sections),
             currentSection = & sectionMap [1];
          position != list_end (& object -> sections);
          position = list_next (position))
     {
         Section * section = (Section *) position;
      
         if (section -> attributes & SECTION_ATTRIBUTE_STRIPPED)
           continue;

         currentSection -> type              = section -> type;
         currentSection -> alignment         = section -> dataAlignment;
         currentSection -> sectionAttributes = section -> attributes & SECTION_ATTRIBUTE_MASK;
         currentSection -> memoryAttributes  = section -> capabilities & SECTION_CAPABILITY_MASK;
         currentSection -> memoryRegion      = (xc_reference) section -> address;
         currentSection -> size              = (xc_reference) section -> dataSize;

         if (section -> name != NULL)
           currentSection -> nameIndex = (xc_reference) namespace_find (& object -> namespace, section -> name) -> offset;

         if (section -> architecture != NULL)
           currentSection -> dataIndex = (xc_reference) namespace_find (& object -> namespace, section -> architecture) -> offset;

         if (zeroFill != 0 &&
             section -> type == SECTION_TYPE_PROGRAM)
           currentSection -> sectionAttributes |= XCOM_SECTION_ATTRIBUTE_CONTENTS;

         ++ currentSection;
     } 

     numSyms = count_symbols (& object -> importedSymbols) + count_symbols (& object -> constantSymbols);
     if (numSyms)
     {
          currentSection -> type = XCOM_SECTION_TYPE_SYMBOLS;
          currentSection -> size = numSyms * sizeof (XComSymbol);

          ++ currentSection;
     }

     for (position = list_begin (& object -> sections),
             targetSection = & sectionMap [1];
          position != list_end (& object -> sections);
          position = list_next (position))
     {
         Section * section = (Section *) position;
         ListIterator relocPos;
         xc_reference numRelocs = 0;

         if (section -> attributes & SECTION_ATTRIBUTE_STRIPPED)
           continue;

         numSyms = count_symbols (& section -> exportedSymbols);
         if (numSyms)
         {
            currentSection -> type       = XCOM_SECTION_TYPE_SYMBOLS;
            currentSection -> dataIndex  = targetSection - sectionMap + 1;
            currentSection -> size       = numSyms * sizeof (XComSymbol);
            
            ++ currentSection;
         }

         targetSection -> sectionAttributes |= XCOM_SECTION_ATTRIBUTE_COMPLETE;

         for (relocPos = list_begin (& section -> relocations);
              relocPos != list_end (& section -> relocations);
              relocPos = list_next (relocPos))
         {
            Relocation * reloc = (Relocation *) relocPos;

            if (reloc -> deferred)
              continue;

            ++ numRelocs;

            targetSection -> sectionAttributes |= XCOM_SECTION_ATTRIBUTE_RELOCATIONS;

            if (reloc -> definition -> symbol -> type != SYMBOL_TYPE_SECTION)
              targetSection -> sectionAttributes &= ~XCOM_SECTION_ATTRIBUTE_COMPLETE;
         }
            
         if (numRelocs)
         {        
            currentSection -> type               = XCOM_SECTION_TYPE_RELOCATIONS;
            currentSection -> sectionAttributes  = XCOM_SECTION_ATTRIBUTE_MANDATORY;
            currentSection -> dataIndex          = targetSection - sectionMap + 1;
            currentSection -> size               = numRelocs * sizeof (XComRelocation);
            
            ++ currentSection;
         }

         ++ targetSection;
     }

     header.sections = currentSection - sectionMap;

     sectionMap -> offset = sizeof (XComHeader) + header.sections * sizeof (XComSection);
     
     currentOffset = sectionMap -> offset + sectionMap -> size;

     for (position = list_begin (& object -> sections),
             currentSection = & sectionMap [1];
          position != list_end (& object -> sections);
          position = list_next (position))
     { 
        Section * section = (Section *) position;
        xc_reference alignment, 
                     alignOffset;

        if (section -> attributes & SECTION_ATTRIBUTE_STRIPPED)
          continue;

        if (section -> type == SECTION_TYPE_PROGRAM)
        { 
           if (zeroFill == 0 &&
               (section -> attributes & SECTION_ATTRIBUTE_CONTENTS) == 0)
             continue;
        
           alignment   = (xc_reference) (1 << currentSection -> alignment);
           alignOffset = currentOffset % alignment;
           if (alignOffset > 0)
             currentOffset += alignment - alignOffset;
        }
        
        currentSection -> offset = currentOffset;

        currentOffset += currentSection -> size;

        ++ currentSection;
     }
    
     while (currentSection < & sectionMap [header.sections])
     {
        currentSection -> offset = currentOffset;

        currentOffset += currentSection -> size;

        ++ currentSection; 
     }

     header.fileSize = currentOffset;
    
     if (fwrite (& header, sizeof (XComHeader), 1, file) != 1)
       xl_error ("Could not write XCOM header");

     if (fwrite (sectionMap, header.sections * sizeof (XComSection), 1, file) != 1)
       xl_error ("Could not write section map for XCOM");

     if (namespace_write (& object -> namespace, file) != 0)
       xl_error ("Could not write string section for XCOM");

     for (position = list_begin (& object -> sections),
             currentSection = & sectionMap [1];
          position != list_end (& object -> sections);
          position = list_next (position))
     {        
         Section * section = (Section *) position;

         if ((section -> attributes & SECTION_ATTRIBUTE_STRIPPED) != 0 || currentSection -> offset <= 0)
           continue;

         if (fseek (file, currentSection -> offset, SEEK_SET) != 0 ||
             section_write (section, file, zeroFill) != 0)
           xl_error ("Could not write data for section %s for XCOM", section -> name);

         ++ currentSection;
     }

     for (position = list_begin (& object -> importedSymbols);
          position != list_end (& object -> importedSymbols);
          position = list_next (position))
     {
         Symbol * symbol = (Symbol *) position;
         XComSymbol currentSymbol;

         if (symbol -> attributes & SYMBOL_ATTRIBUTE_STRIPPED)
           continue;

         currentSymbol.type       = symbol -> type;
         currentSymbol.attributes = symbol -> attributes & SYMBOL_ATTRIBUTE_MASK;
         currentSymbol.size       = 0;
         currentSymbol.offset     = (xc_reference) symbol -> value;
         currentSymbol.nameIndex  = (xc_reference) symbol -> definition -> offset;

         if (fwrite (& currentSymbol, sizeof (XComSymbol), 1, file) != 1)
           xl_error ("Could not write imported symbol %s for XCOM", symbol -> definition -> name);
     }

     for (position = list_begin (& object -> constantSymbols);
          position != list_end (& object -> constantSymbols);
          position = list_next (position))
     {
         Symbol * symbol = (Symbol *) position;
         XComSymbol currentSymbol;

         if (symbol -> attributes & SYMBOL_ATTRIBUTE_STRIPPED)
            continue;

         currentSymbol.type       = symbol -> type;
         currentSymbol.attributes = symbol -> attributes & SYMBOL_ATTRIBUTE_MASK;
         currentSymbol.size       = 0;
         currentSymbol.offset     = (xc_reference) symbol -> value;
         currentSymbol.nameIndex  = (xc_reference) symbol -> definition -> offset;

         if (fwrite (& currentSymbol, sizeof (XComSymbol), 1, file) != 1)
           xl_error ("Could not write constant symbol %s for XCOM", symbol -> definition -> name);
     }

     for (position = list_begin (& object -> sections);
          position != list_end (& object -> sections);
          position = list_next (position))
     {
         Section * section = (Section *) position;
         ListIterator symbolPos, relocPos;

         if (section -> attributes & SECTION_ATTRIBUTE_STRIPPED)
           continue;

         for (symbolPos = list_begin (& section -> exportedSymbols);
              symbolPos != list_end (& section -> exportedSymbols);
              symbolPos = list_next (symbolPos))
         {
             Symbol * symbol = (Symbol *) symbolPos;
             XComSymbol currentSymbol;

             if (symbol -> attributes & SYMBOL_ATTRIBUTE_STRIPPED)
               continue;

             currentSymbol.type       = symbol -> type;
             currentSymbol.attributes = symbol -> attributes & SYMBOL_ATTRIBUTE_MASK;
             currentSymbol.size       = 0;
             currentSymbol.offset     = (xc_reference) symbol -> value;
             currentSymbol.nameIndex  = (xc_reference) symbol -> definition -> offset;

             if (fwrite (& currentSymbol, sizeof (XComSymbol), 1, file) != 1)
               xl_error ("Could not write exported symbol %s for XCOM", symbol -> definition -> name);
         }

         for (relocPos = list_begin (& section -> relocations);
              relocPos != list_end (& section -> relocations);
              relocPos = list_next (relocPos))
         {
             Relocation * reloc = (Relocation *) relocPos;
             XComRelocation currentReloc;

             if (reloc -> deferred)
               continue;

             currentReloc.relocationType = reloc -> type;
             currentReloc.symbolType     = reloc -> definition -> symbol -> type;
             currentReloc.size           = 0;
             currentReloc.offset         = (xc_reference) reloc -> offset;
             currentReloc.nameIndex      = (xc_reference) reloc -> definition -> offset;

             if (fwrite (& currentReloc, sizeof (XComRelocation), 1, file) != 1)
               xl_error ("Could not write relocation against %s in section %s for XCOM", 
                         reloc -> definition -> name,
                         section -> name);
         }        
     }

     free (sectionMap);

     return 0;
}
                 
